"""Scripts to commit and push GitHub repos.

In development. Pushing does not work yet.

The problem is how to use GitHub authentication.

"""

import subprocess
import os

def run_command(command, cwd, verbose=True):
    """Run a command line command and print the output.

    Args:
        command (list): a list containing the elements of the command,
                        e.g., ["git", "commit", "-m", msg]
        cwd (str): path to folder where command needs to be carried out
        verbose (bool): if verbose is set to True, output is printed

    See https://www.endpoint.com/blog/2015/01/28/getting-realtime-output-using-python
    """
    process = subprocess.Popen(command, cwd=cwd,
                               stdout=subprocess.PIPE)
    print("process started")
    while True:
        output = process.stdout.readline()
        if process.poll() is not None:
            print(process.poll())
            break
        if verbose:
            if output:
                print(output.strip())
    print("finished")
    rc = process.poll()
    return rc

def run_command(command, cwd, verbose=True):
    """Run a command line command and print the output.

    Args:
        command (list): a list containing the elements of the command,
                        e.g., ["git", "commit", "-m", msg]
        cwd (str): path to folder where command needs to be carried out
        verbose (bool): if verbose is set to True, output is printed
    """
    output = subprocess.check_output(command, cwd=cwd)
    if verbose:
        print(output.decode().strip())
    return output.decode().strip()


def get_commit_msg(msg=None):
    """Define a message for the commit.

    Args:
        msg (str): user-defined message. If no msg is provided,
            the user will be given the choice between
            a default commit message and a user-defined message.
    """
    if msg:
        return msg
    else:
        default_msg = "Updated repo"
        print("Use default commit message?")
        print("    '{}'".format(default_msg))
        resp = input("Y/N: ")
        if resp.upper() == "N":
            msg = input("Write commit message: ")
        else:
            msg = default_msg
        return msg

    
def is_git_repo(pth="."):
    """Check whether the folder at pth is a git repo.

    See https://stackoverflow.com/a/24584384

    Returns:
        bool: True if pth is a git repo, False if it isn't.
    """
    return subprocess.call(["git", "-C", pth, 'status'],
                           stderr=subprocess.STDOUT,
                           stdout = open(os.devnull, 'w')) == 0
    

def commit_repo(repo_pth, add=["."], msg=None):
    """Add changed files and commit.

    Args:
        repo_pth (str): path to the repository to be committed
        add (list): a list of file patterns to be added for commit
        msg (str): a commit message. If None, user will be asked
            whether a default message should be used,
            or else to provide one.
    """
    if not msg:
        msg = get_commit_msg(msg)
    if is_git_repo(repo_pth):
        for it in add:
            print("GIT ADD", it)
            run_command(["git", "add", it], cwd=repo_pth)
##            p = subprocess.Popen(["git", "add", it],
##                                 cwd=repo_pth)
##            p.wait()
        print("GIT COMMIT -M '{}'".format(msg))
        run_command(["git", "commit", "-m", msg], cwd=repo_pth)
##        p = subprocess.Popen(["git", "commit", "-m", msg],
##                             cwd=repo_pth)
        print("Repo {} committed successfully".format(repo_pth))
    else:
        print("Aborted: folder {} is not a git repo".format(repo_pth))

def commit_all_repos_in_folder(folder, add=["."], msg=None):
    """Commit changes in all repos in a folder.

    Args:
        folder (str): path to the folder containing the repos
        add (list): a list of file patterns to be added for commit
        msg (str): a commit message. If None, user will be asked
            whether a default message should be used,
            or else to provide one.
    """
    if not msg:
        msg = get_commit_msg(msg)
    for repo in os.listdir(folder):
        repo_pth = os.path.join(folder, repo)
        commit_repo(repo_pth, add, msg)

def get_remote_url(repo_pth):
    """Get the remote url of a repo."""
    if is_git_repo(repo_pth):
        print("GIT CONFIG --GET REMOTE.ORIGIN.URL")
        return run_command(["git", "config", "--get", "remote.origin.url"],
                           cwd=repo_pth)
    else:
        print("Aborted: folder {} is not a git repo".format(repo))

def push(repo_pth, branch="master"):
    """Push a repo to the remote.
    THIS FUNCTION DOES NOT WORK !!!!!"""
    if is_git_repo(repo_pth):
        try:
            print("GIT PUSH -U ORIGIN", branch)
            run_command(["git", "push", "-u", "origin", branch], cwd=repo_pth)
        except:
            print("failed")
            try:
                remote = get_remote_url(repo_pth)
            except:
                print(repo_pth, ": no remote defined!")
                return
            print("GIT PUSH", remote, branch)
            run_command(["git", "push", "-u", remote, branch], cwd=repo_pth)
            print("Repo {} pushed to remote".format(repo))
    else:
        print("Aborted: folder {} is not a git repo".format(repo))
            
def push_all_repos_in_folder(folder, origin="origin", branch="master"):
    """Push all repos in a folder to the remote."""
    for repo in os.listdir(folder):
        repo_pth = os.path.join(folder, repo)
        push(repo_pth, origin="origin", branch="master")

            
def add_origin(repo_pth, user, token, remote_url, origin="origin"):
    """Add a remote origin to a repo.

    Args:
        repo_pth (str): path to the repo
        user (str): user name of the user
        token (str): GitHub access token;
            if None, user will be prompted to provide it.
        remote_url (str): url of the remote git
        origin (str): name of the remote.

    See https://stackoverflow.com/questions/18935539/authenticate-with-github-using-a-token
    """
    if token == None:
        token = input("GitHub access token: ")
    #r = remote_url.replace("//", "//{}:{}@".format(user, token))
    r = remote_url.replace("//", "//{}@".format(token))
    try:
        print("GIT REMOTE SET-URL", origin, r)
        run_command(["git", "remote", "set-url", origin, r],
                    cwd=repo_pth)
    except:
        print("failed; try:")
        print("GIT REMOTE ADD", origin, r)
        run_command(["git", "remote", "add", origin, r],
                    cwd=repo_pth)

    
if __name__ == "__main__":

    token_fp = r"D:\London\OpenITI\python_library\test_git_repo\GitHub personalAccessTokenReadOnly.txt"
    try:
        with open(token_fp,
                  mode="r", encoding="utf-8") as file:
            token = file.read().strip()
    except:
        token = None # you will be prompted to insert the token manually
    print("token:", token)

    folder = r"D:\London\OpenITI\python_library\test_git_repo"
    print(is_git_repo(folder))


    remote_url = "https://github.com/pverkind/test_git_repo.git"
    add_origin(folder, "pverkind", token, remote_url)



    commit_repo(folder, add=["README.md", ".gitignore"],
                msg="Automatic commit")
    run_command(["git", "status"], cwd=folder)

    remote = run_command(["git", "config", "--get", "remote.origin.url"],
                         cwd=folder)
    print(remote)

    push(folder, branch="master")
    ##msg = "Updated yml files during automatic metadata creation."
    ##commit_all_repos_in_folder(folder, msg)



"""Converter that converts HTML files from the Ghbook library
to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single html file.
* convert_files_in_folder: convert all html files in a given folder

Usage examples:
    >>> from html_converter_Ghbook import convert_file
    >>> folder = r"test/Ghbook/"
    >>> convert_file(folder+"10584.html", dest_fp=folder+"converted/10584")
    >>> from html_converter_Ghbook import convert_files_in_folder
    >>> convert_files_in_folder(folder, dest_folder=folder+"converted")  

Both functions use the GhbookHtmlConverter class to do the heavy lifting.
The GhbookHtmlConverter is a subclass of GenericHtmlConverter,
which in its turn inherits many functions from the GenericConverter.

GenericConverter
    \_ GenericHtmlConverter
            \_ EShiaHtmlConverter
            \_ GhbookHtmlConverter
            \_ ...

Overview of the methods of these classes:
(methods of GenericConverter are inherited by GenericHtmlConverter;
and methods of GenericHtmlConverter are inherited by GhbookHtmlConverter.
Methods of the superclass with the same name
in the subclass are overwritten by the latter)

================================== ========================== ==========================
GenericConverter                   GenericHtmlConverter       GhbookHtmlConverter
================================== ========================== ==========================
__init__                           __init__                   (inherited)
convert_files_in_folder            (inherited)                (inherited)
convert file                       (inherited)                (inherited)
make_dest_fp                       (inherited - generic!)     (inherited - generic!)
get_metadata (dummy)               (inherited - dummy!)       get_metadata
get_data                           get_data                   (inherited)
pre_process                        (inherited)                pre_process
add_page_numbers (dummy)           (inherited - dummy!)       add_page_numbers
add_structural_annotations (dummy) add_structural_annotations add_structural_annotations
remove_notes (dummy)               remove_notes               remove_notes
reflow                             (inherited)                (inherited)
add_milestones (dummy)             (inherited - dummy!)       (inherited - dummy!)
post_process                       (inherited - generic!)     post_process
compose                            (inherited)                (inherited)
save_file                          (inherited)                (inherited)
                                   inspect_tags_in_html       (inherited)
                                   inspect_tags_in_folder     (inherited)
                                   find_example_of_tag        (inherited)
================================== ========================== ==========================

The GhbookHtmlConverter's add_structural_annotations method uses html2md_Ghbook,
an adaptation of the generic html2md (based on markdownify)
to convert the html tags to OpenITI annotations. 

Examples:
    >>> from html_converter_Ghbook import GhbookHtmlConverter
    >>> conv = GhbookHtmlConverter()
    >>> conv.dest_folder = r"test/Ghbook/converted"
    >>> conv.VERBOSE = False
    >>> folder = r"test/Ghbook/"
    >>> conv.convert_file(folder+"10584.html")
    >>> conv.convert_files_in_folder(folder, ["html"])
"""

from bs4 import BeautifulSoup
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.html_converter_generic import GenericHtmlConverter
from openiti.new_books.convert.helper import html2md_Ghbook


def convert_file(fp, dest_fp=None):
    """Convert one file to OpenITI format.

    Args:
        source_fp (str): path to the file that must be converted.
        dest_fp (str): path to the converted file. Defaults to None
            (in which case, the converted folder will be put in a folder
            named "converted" in the same folder as the source_fp)

    Returns:
        None
    """
    conv = GhbookHtmlConverter() 
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, dest_folder=None,
                            extensions=["html"], exclude_extensions=["yml"],
                            fn_regex=None):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    conv = GhbookHtmlConverter()
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################


class GhbookHtmlConverter(GenericHtmlConverter):

    def pre_process(self, text):
        """Remove superfluous elements from the html file before processing."""

        def remove_html_elements(soup, tag_name, class_=None, contains_str=None):
            """Remove all html elements with tag `tag` and class `class_` \
            if they contain `contains_str`
            
            Args:
                soup (BeautifulSoup object): BeautifulSoup representation
                    of the html file
                tag_name (str): name of the tag that needs to be removed
                    (e.g., "p", "div", "span"). 
                class_ (str): class of the tag that needs to be removed.
                    Defaults to None. If None, all `tag_name` elements
                    will be removed, regardless of their class.
                contains_str (str): defaults to None. If not None,
                    `tag_name` tags will only be removed if the text within them
                    contains the `contains_str` string.
            """
            if class_:
                elements = soup.find_all(tag_name, class_=class_)
            else:
                elements = soup.find_all(tag_name)
            for el in elements:
                if contains_str:
                    if contains_str in el.text:
                        el.extract()
                else:
                    el.extract()
        
        text = super().pre_process(text)
        
        # attach separated wa- and a- prefixes to the following word: 
        text = re.sub(r"\b([وأ])[\s~]+", r"\1", text)

        # remove superfluous html elements: 
        soup = BeautifulSoup(text)
        remove_html_elements(soup, "STYLE")
        remove_html_elements(soup, "TITLE")

        for fn_line in soup.find_all("hr", class_="content_hr"):
            fn_line.insert_after("FOOTNOTES")
        text = soup.prettify()
        
        return text


    def add_page_numbers(self, text, source_fp):
        """Convert the page numbers in the text into OpenITI mARkdown format

        In Ghbook texts, the page numbers are in the page header
        (<div class="PageHead">).
        Volume numbers are not mentioned in the html files,
        but every volume is a different html file
        and volume numbers should be marked in the file names as VOLxxx.

        The script gets the volume number from the file name
        and the page number from the page header,
        joins these together in the OpenITI page number format PageVxxPxxx
        and adds this into the html at the end of the page.

        It also deletes the page header after extracting the page number.
        """

        # try to get the volume number from the filename:
        try:
            vol_no = int(re.findall("VOL(\d+)", source_fp)[0])
            vol_no = "PageV{:02d}P{}".format(vol_no, "{:03d}")
        except:
            vol_no = "PageV01P{:03d}"

        # add the page number
        soup = BeautifulSoup(text)
        for span in soup.find_all("SPAN", class_="content_text"):
            span_text = span.text.strip()
            if span_text.startswith("ص:"):
                page_no = re.findall("\d+", span_text)[0]
                page_no = vol_no.format(int(page_no))
                span.insert_after(page_no)
                span.extract()
        return soup.prettify()


    def remove_notes(self, text):
        """Remove footnotes from text, and format them as endnotes.

        Footnotes in Ghbook html files are below a horizontal line
        (<HR class=content_hr>), located just below the page number
        (<P class=content_paragraph><SPAN class=content_text>ص:45</SPAN></P></DIV></DIV>)
        each footnote in a <DIV id=content_note_PAGE_NOTENUMBER class=content_note>(footnote text)</DIV>
        
        This function extracts the footnotes from the texts
        and turns them into endnotes.
        The markers that indicate the location of the notes
        within the text are not removed.
        """
        split_text = re.split("(PageV\d+P\d+)", text)
        text = []
        footnotes = []
        for i, t in enumerate(split_text):
            if re.match("PageV\d+P\d+", t):
                text.append(t)
            else:
                notes = re.findall("content_note>([^<]+)", t)
                if notes:
                    try:
                        notes = "\n".join(notes) + "\n" +  split_text[i-1] + "\n\n"
                    except:
                        notes = "\n".join(notes) + "\n" + "PageV00P000\n\n"
                    footnotes.append(notes)
                    
                text.append(re.sub("<DIV [^>]+? class=content_note>[^<]+?</DIV>", "", t))

        text = "\n\n".join(text)
        notes = "\n\n".join(footnotes)
        notes = re.sub("\n+#* *\n+", "\n\n", notes)
        notes = self.endnote_splitter + notes
        return text, notes 
                    

    def add_structural_annotations(self, html):
        """Convert html to mARkdown text using a html2md converter."""
        text = html2md_Ghbook.markdownify(html)
        return text             


    def post_process(self, text):
        """Deal with formatting probles introduced during the conversion process."""
        text = super().post_process(text)

        # remove page numbers of empty pages:
        text = re.sub("(PageV\d+P\d+)\s*PageV\d+P\d+", r"\1", text)

        # remove empty paragraphs:
        text = re.sub(r"[\r\n]+# *[\r\n]+", "\n", text)

        # adjust spacing after closing brackets and punctuation:
        fmt = ")»،؛:.!؟\-"
        fmt2 = fmt + "\d\s"
        text = re.sub("([{}]+)([^{}])".format(fmt, fmt2), r"\1 \2", text)
        text = re.sub("\) ([{}])".format(fmt), r")\1", text)

        # adjust spacing before opening brackets:
        text = re.sub("(\w)([(«])", r"\1 \2", text)

        # remove superfluous new lines before a new paragraph/page number
        text = re.sub("[\r\n]+(# |Page)", r"\n\1", text)

        return text


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    input("Passed all tests. Continue?")





# inserting milestones for splitting texts into chunks of the same size

import re, os
#import helper.funcs as hlp
#import zfunc
from openiti.helper import funcs, ara

stopper = 1


def reformatting(path_full, target_folder):
    file_name = path_full.split("/")[-1]
    file_id = "_".join(file_name.split(".")[:2])
    print("Processing:")

    target_path = target_folder + file_id
    print("\t" + target_path)

    # run generation
    if os.path.exists(target_path):
        count = 0
        print("\t\t! The file has been processed")
    else:
        print("Converting:")
        print("\n" + file_name)
        print("\ninto:\t" + file_id)
        with open(path_full, "r", encoding="utf8") as f1:
            data = f1.read()
            #data = data.split(hlp.funcs.splitter)[1]
            data = data.split(funcs.splitter)[1]

            data = data.replace("\n~~", " ")
            # data = data.replace("#", "(@@)")
            data = data.replace("Milestone300", " ")

            data = re.sub(" +", " ", data)

            #data = hlp.ara.normalize_ara_extra_light(data)
            data = ara.normalize_ara_extra_light(data)
            data = re.sub("Page\w+|[#\|\-\%\d]", "", data)

            with open(target_path, "w", encoding="utf8") as ft:
                ft.write(data)

            count = 1

    return count


def load_meta_data():
    # TODO: interactive path to metedata file
    meta_file = "../Annotation/OpenITI_metadata_light.csv"
    meta = {}
    with open(meta_file, "r", encoding="utf8") as f1:
        data = f1.read().split("\n")

        for d in data[1:]:
            da = d.split("\t")
            if da[5] == "pri":
                meta[da[0].strip()] = da

    return meta


# process all texts in OpenITI
def process_all(folder, target_path):
    meta_data = load_meta_data()
    meta_stylo = []

    print()
    print("Generating i.stylo corpus...")
    print()

    count = 1
    not_in_meta_list = []

    for root, dirs, files in os.walk(folder):
        #dirs[:] = [d for d in dirs if d not in zfunc.exclude]
        dirs[:] = [d for d in dirs if d not in funcs.exclude]

        for file in files:
            if re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))?$", file):
                if "DhaylTabaqatHanabila" in file:
                    input(file)
                if file in meta_data:
                    path_full = os.path.join(root, file)
                    file_stylo = "_".join(file.split(".")[:2])
                    meta_stylo.append(file_stylo + "\t" + file + "\t%d" % int(file_stylo[:4]))

                    if "DhaylTabaqatHanabila" in file:
                        print(path_full)
                        input(file)

                    if "" in path_full:
                        count += reformatting(path_full)
                        if count % stopper == 0:
                            print()
                            print("=============" * 2)
                            print("Processed: %d" % count)
                            print("=============" * 2)
                            print()

                    if count == stopper:
                        print()

    with open(target_path.replace("_Temp", "") + "meta_stylo", "w", encoding="utf8") as ft:
        ft.write("\n".join(meta_stylo))

    print("Not in META")
    print("\n".join(not_in_meta_list))



"""Convert html to Markdown.

This program is an adaptation of python-markdownify
(https://github.com/matthewwithanm/python-markdownify)
to output OpenITI mARkdown.
It also adds methods for tables and images,
and a post-processing method.

The main componant of this script is the `MarkdownConverter` class,
which contains a basic procedure for converting html, tag by tag,
(`MarkdownConverter.convert`)
and methods for converting specific html tags to mARkdown
(`MarkdownConverter.convert_a`, `MarkdownConverter.convert_img`, ...).

The easiest way to use the MarkdownConverter is to use the
`markdownify` function, which calls the `convert` method of
the `MarkdownConverter` class.

You can use this class as a base class and subclass it
to add methods, adapt the post-processing method etc.

E.g.:
    class Hindawi_converter(html2md.MarkdownConverter):
        def post_process_md(self, text):
            text = super().post_process_md(text)
            # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
            text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
            # replace placeholders for spaces in tables:
            text = re.sub("ç", " ", text)
            return text


Examples (doctests):

    Headings: h1

    >>> import html2md
    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h)
    '\\n\\n### | abc\\n\\n'

    NB: heading style is OpenITI mARkdown style by default,
        but can be set to other styles as well:

    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h, md_style=UNDERLINED)
    '\\n\\nabc\\n===\\n\\n'

    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h, md_style=ATX)
    '\\n\\n# abc\\n\\n'

    Paragraphs (<p>):

    >>> h = "<p>abc</p>"
    >>> html2md.markdownify(h)
    '\\n\\n# abc\\n\\n'

    >>> h = "<p>abc</p>"
    >>> html2md.markdownify(h, md_style=ATX)
    '\\n\\nabc\\n\\n'


    Divs without class or with an unsupported class are stripped:

    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'

    >>> h = 'abc\
             <div class="unknown_div_class">def</div>\
             ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'


    Spans without class or with an unsupported class are stripped:

    >>> h = 'abc <span>def</span> ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'

    >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'


    Links:

    >>> h = '<a href="a/b/c">abc</a>'
    >>> html2md.markdownify(h)
    '[abc](a/b/c)'


    Unordered lists:

    >>> h = '<ul><li>item1</li><li>item2</li></ul>'
    >>> html2md.markdownify(h)
    '\\n* item1\\n* item2\\n\\n'

    Ordered lists:

    >>> h = '<ol><li>item1</li><li>item2</li></ol>'
    >>> html2md.markdownify(h)
    '\\n1. item1\\n2. item2\\n\\n'

    Nested lists:

    >>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
    >>> html2md.markdownify(h)
    '\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'

    Italics (<i> and <em> tags):

    >>> h = 'abc <em>def</em> ghi'
    >>> html2md.markdownify(h)
    'abc *def* ghi'

    >>> h = 'abc <i>def</i> ghi'
    >>> html2md.markdownify(h)
    'abc *def* ghi'


    Bold (<b> and <strong> tags):

    >>> h = 'abc <b>def</b> ghi'
    >>> html2md.markdownify(h)
    'abc **def** ghi'

    >>> h = 'abc <strong>def</strong> ghi'
    >>> html2md.markdownify(h)
    'abc **def** ghi'

    Tables:

    >>> h = '\
    <table>\
      <tr>\
        <th>th1aaa</th><th>th2</th>\
      </tr>\
      <tr>\
        <td>td1</td><td>td2</td>\
      </tr>\
    </table>'
    >>> html2md.markdownify(h)
    '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'

    # i.e.,
    # | th1aaa | th2 |
    # | td1    | td2 |
"""

from bs4 import BeautifulSoup, NavigableString, Comment
import re
#import six


convert_heading_re = re.compile(r'convert_h(\d+)')
line_beginning_re = re.compile(r'^', re.MULTILINE)
whitespace_re = re.compile(r'[\r\n\s\t ]+')
FRAGMENT_ID = '__MARKDOWNIFY_WRAPPER__'
wrapped = '<div id="%s">%%s</div>' % FRAGMENT_ID



# Heading styles
ATX = 'atx'
ATX_CLOSED = 'atx_closed'
UNDERLINED = 'underlined'
OPENITI = 'OpenITI'
SETEXT = UNDERLINED


def escape(text):
##    if not text:
##        return ''
##    return text.replace('_', r'\_')
    return text


def _todict(obj):
    return dict((k, getattr(obj, k)) for k in dir(obj) if not k.startswith('_'))


class MarkdownConverter(object):
    class DefaultOptions:
        strip = None
        convert = None
        autolinks = True
        md_style = OPENITI
        bullets = '*+-'  # An iterable of bullet types.
        image_link_regex = ""  # e.g., "images/books"
        image_folder = "img"

    class Options(DefaultOptions):
        pass

    def __init__(self, **options):
        # Create an options dictionary. Use DefaultOptions as a base so that
        # it doesn't have to be extended.
        self.options = _todict(self.DefaultOptions)
        self.options.update(_todict(self.Options))
        self.options.update(options)
        #print("options", self.options)
        if self.options['strip'] is not None and self.options['convert'] is not None:
            raise ValueError('You may specify either tags to strip or tags to'
                             ' convert, but not both.')


    def convert(self, html):
        """Convert html to markdown.

        # We want to take advantage of the html5 parsing, but we don't actually
        # want a full document. Therefore, we'll mark our fragment with an id,
        # create the document, and extract the element with the id.
        """

        html = wrapped % html
        soup = BeautifulSoup(html, 'html.parser')
        if 'strip' in self.options and self.options["strip"]:
            for tag in self.options["strip"]:
                [t.decompose() for t in soup.find_all(tag)]
        text = self.process_tag(soup.find(id=FRAGMENT_ID), children_only=True)

##        # post-processing: remove unneeded blank lines and spaces:
##
##        # remove leading and trailing spaces in lines:
##        text = re.sub(r" *(\n+) *", r"\1", text)
##        # remove unwanted additional spaces and lines:
##        text = re.sub(r"\n{3,}", r"\n\n", text)
##        text = re.sub(r" +", r" ", text)
##        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
##        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
##        # replace placeholders for spaces in tables:
##        text = re.sub("ç", " ", text)
##        return text
        return self.post_process_md(text)

    def process_tag(self, node, children_only=False):
        """Process each tag and its children.
        """
        #print(node.name, node.attrs)
        text = ''

        # Convert the children first
        for el in node.children:

            if isinstance(el, NavigableString):
                if not isinstance(el, Comment):  # remove html comments
                    #text += self.process_text(six.text_type(el))
                    text += self.process_text(str(el))
            else:
                #print(text[-50:])
                text += self.process_tag(el)


        if not children_only:
            convert_fn = getattr(self, 'convert_%s' % node.name, None)
            if convert_fn and self.should_convert_tag(node.name):
                #print("text is now:")
                #print(text[:50])
                #input(text)
                text = convert_fn(node, text)
        return text

    def process_text(self, text):
        #return escape(whitespace_re.sub(' ', text or ''))
        if whitespace_re.sub('', text or ''):
            return whitespace_re.sub(' ', text)
        return ""

    def fill_out_columns(self, match):
        """Find the longest cell in a column; add spaces to shorter columns."""

        # split the table into a list of lists:

        table = match.group(1)
        #print(table)
        if "|--" in table:
            header = True
        else:
            header = False
        table = table.splitlines()
        if header:
            del table[1]  # remove horizontal lines separating header from body
        table = [(row.split("|"))[1:] for row in table]  # remove empty string

        # get length of longest cell per column:

        column_length = dict()
        for row in table:
            for i, cell in enumerate(row):
                try:
                    if len(cell) > column_length[i]:
                        column_length[i] = len(cell)
                except:
                    column_length[i] = len(cell)

        # fill out shorter cells:

        new_table = []
        for row in table:
            new_row = []
            for i, cell in enumerate(row):
                new_cell = cell + " "*(column_length[i]-len(cell))
                new_cell = new_cell + (" " * len(re.findall("لا|لأ|لإ", new_cell)))
                #combining = re.findall("[ؐ-ًؚ-ٰٟۖ-ۜ۟-۪ۤۧۨ-ۭ]", new_cell)
                #new_cell = re.sub(" {%s}" % len(combining), "", new_cell)
                new_row.append(new_cell)
            new_table.append(new_row)

        # re-build the table:

        new_table = [" | ".join(row).strip() for row in new_table]
        if header:
            lines = " | ".join(["-" * v for k,v in column_length.items()])
            new_table = [new_table[0], lines.strip()]+new_table[1:]
        new_table = "\n| ".join(new_table)
        new_table = "\n\n| {}\n\n".format(new_table)

        #print("new_table: ")
        #print(new_table)
        return new_table


    def post_process_named_entities(self, match):
        """Reformat named entity matches to mARkdown named entity standard.

        Named entities should be marked with @TAG@ tags in the converter
        (3 capital letters between ampersands), and end with a new line.
        This post-processing step then converts these temporary tags
        into OpenITI mARkdown format @TAG\d\d+:
        
        * The first number after the @QUR tag refers to the number of letters
          following the tag that do not belong to the named entity
          (in this automatic step, this number will always be set to 0);
        * the following number(s) refer(s) to the length of the entity in tokens

        Examples:
            >>> import html2md
            >>> conv = html2md.MarkdownConverter()
            >>> conv.post_process_md("abc @QUR@ def ghi\\njkl")
            'abc @QUR02 def ghi jkl'
            >>> conv.post_process_md("abc @QUR@ def ghi\\n~~jkl\\nmno")
            'abc @QUR03 def ghi\\n~~jkl mno'
        """

        foll_char = match.group(3)
        entity = match.group(2)
        ent_words = len(re.findall("[\n\r ،؛:.!؟\-]+", entity)) + 1
        code = match.group(1)
        if code.startswith("QUR"):
            return "@{}@0{} {} {}".format(code, ent_words, entity, foll_char)
        return "@{}0{} {} {}".format(code, ent_words, entity, foll_char)


    def post_process_md(self, text):
        """Post-processing operation to improve formatting of converted text."""
        # post-process named entity tags:
        text = re.sub("@([A-Z.\d]+)@ +(.+?)\n([^~]|Z)",
                      self.post_process_named_entities, text,
                      flags=re.DOTALL)
        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)

        # remove leading and trailing spaces in lines:
        text = re.sub(r" *(\n+) *", r"\1", text)
        # remove unwanted additional spaces and lines:
        text = re.sub(r"\n{3,}", r"\n\n", text)
        text = re.sub(r" +", r" ", text)
        # fill out columns in tables:
        #text = re.sub("\n\n(\|.+?)\n\n", self.fill_out_columns, text, flags=re.DOTALL)
        return text

    def __getattr__(self, attr):
        # Handle headings
        m = convert_heading_re.match(attr)
        if m:
            n = int(m.group(1))

            def convert_tag(el, text):
                return self.convert_hn(n, el, text)

            convert_tag.__name__ = 'convert_h%s' % n
            setattr(self, convert_tag.__name__, convert_tag)
            return convert_tag

        raise AttributeError(attr)

    def should_convert_tag(self, tag):
        """Check whether a tag should be converted or simply stripped"""
        tag = tag.lower()
        strip = self.options['strip']
        convert = self.options['convert']
        if strip is not None:
            return tag not in strip
        elif convert is not None:
            return tag in convert
        else:
            return True

    def indent(self, text, level):
        """Add tab indentation before text."""
        return line_beginning_re.sub('\t' * level, text) if text else ''

    def create_underline_line(self, text, pad_char):
        """Create a sequence of pad_char characters the same lenght as text."""
        return pad_char * len(text) if text else ''

    def underline(self, text, pad_char):
        """Underline text with pad_char characters (-, =, or +).

        Args:
            text (str): the text within a tag, to be underlined
            pad_char (str): the character used for the line (-, =, or +)

        Returns:
            an underlined line of text

        Example:
            >>> import html2md
            >>> html2md.MarkdownConverter().underline("123456789", "=")
            '123456789\\n========='
            >>> html2md.MarkdownConverter().underline("123456789  ", "=")
            '123456789\\n========='
        """
        text = (text or '').rstrip()
        line = self.create_underline_line(text, pad_char)
        return "{}\n{}".format(text, line)



    # Conversion functions for specific tags (in alphabetic order):

    def convert_a(self, el, text):
        """Convert html links to markdown-style links.

        Example:
            >>> import html2md
            >>> h = '<a href="a/b/c">abc</a>'
            >>> html2md.markdownify(h)
            '[abc](a/b/c)'
        """
        href = el.get('href')
        title = el.get('title')
        if self.options['autolinks'] and text == href and not title:
            # Shortcut syntax
            return '<%s>' % href
        title_part = ' "%s"' % title.replace('"', r'\"') if title else ''
        return '[%s](%s%s)' % (text or '', href, title_part) if href else text or ''

    def convert_b(self, el, text):
        """Convert <b> tags into markdown formatting"""
        return self.convert_strong(el, text)

    def convert_blockquote(self, el, text):
        """Convert <blockquote> tags into markdown formatting"""
        return '\n' + line_beginning_re.sub('> ', text) if text else ''

    def convert_br(self, el, text):
        """Convert <br/> tags into newline characters."""
        return '  \n'

    def convert_em(self, el, text):
        """convert <em> (italics) tags into markdown formatting."""
        return '*%s*' % text if text else ''

    def convert_hn(self, n, el, text):
        """Convert html headings (<h1>, <h2>, etc. into markdown formatting."""
        style = self.options['md_style']
        text = text.rstrip()
        if style == OPENITI:
            return '\n\n### %s %s\n\n' % ("|"*n, text)
        if style == UNDERLINED and n <= 2:
            line = '=' if n == 1 else '-'
            return '\n\n%s\n\n' % self.underline(text, line)
        hashes = '#' * n
        if style == ATX_CLOSED:
            return '\n\n%s %s %s\n\n' % (hashes, text, hashes)
        return '\n\n%s %s\n\n' % (hashes, text)

    def convert_i(self, el, text):
        """convert <i> (italics) tags into markdown formatting."""
        return self.convert_em(el, text)

    def convert_img(self, el, text):
        """Convert <img> tags into markdown-style links to image files.
        
        Examples:
            >>> import html2md
            >>> h = '<div><img class="figure" src="../Images/figure1.png" /></div>'
            >>> html2md.markdownify(h)
            '![](../Images/figure1.png)'
            
            >>> html2md.markdownify(h, image_link_regex="../Images", image_folder="img")
            '![](img/figure1.png)'
        """

        alt = el.attrs.get('alt', None) or ''
        src = el.attrs.get('src', None) or ''
        title = el.attrs.get('title', None) or ''
        title_part = ' "%s"' % title.replace('"', r'\"') if title else ''

        # replace the online address with the local path to the image:
        if self.options['image_link_regex']:
            src = re.sub(self.options['image_link_regex'],
                         self.options['image_folder'], src)

        return '![%s](%s%s)' % (alt, src, title_part)

    def convert_list(self, el, text):
        """Convert ordered and unordered html lists (<ul> and <ol> tags).

        Examples:
            # unordered lists:
            
            >>> import html2md
            >>> h = '<ul><li>item1</li><li>item2</li></ul>'
            >>> html2md.markdownify(h)
            '\\n* item1\\n* item2\\n\\n'

            # ordered lists:
            
            >>> import html2md
            >>> h = '<ol><li>item1</li><li>item2</li></ol>'
            >>> html2md.markdownify(h)
            '\\n1. item1\\n2. item2\\n\\n'

            # nested lists::
            
                ###### TEST FAILS FOR UNKNOWN REASONS
                ##>>> import html2md
                ##>>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
                ##>>> html2md.markdownify(h)
                ##'\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'
        """
        nested = False
        while el:
            if el.name == 'li':
                nested = True
                break
            el = el.parent
        if nested:
            text = '\n' + self.indent(text, 1)
        return '\n' + text + '\n'


    def convert_li(self, el, text):
        """Convert list element tags <li>."""
        parent = el.parent
        if parent is not None and parent.name == 'ol':
            bullet = '%s.' % (parent.index(el) + 1)
        else:
            depth = -1
            while el:
                if el.name == 'ul':
                    depth += 1
                el = el.parent
            bullets = self.options['bullets']
            bullet = bullets[depth % len(bullets)]
        return '%s %s\n' % (bullet, text or '')

    convert_ol = convert_list


    def convert_p(self, el, text):
        """Converts <p> tags.

        Examples:
            >>> import html2md
            >>> h = "<p>abc</p>"
            >>> html2md.markdownify(h)
            '\\n\\n# abc\\n\\n'

            >>> h = "<p>abc</p>"
            >>> html2md.markdownify(h, md_style=ATX)
            '\\n\\nabc\\n\\n'

            >>> h = "<p></p>"
            >>> html2md.markdownify(h, md_style=ATX)
            ''
        """
        if self.options['md_style'] == OPENITI:
            return '\n\n# %s\n\n' % text if text else ''
        else:
            return '\n\n%s\n\n' % text if text else ''


    def convert_table(self, el, text):
        """Wrap tables between double new lines.

        NB: conversion of the tables takes place on the tr level."""
        #print("CONVERTING TABLE")
        return '\n\n' + text + '\n\n'

    def convert_tr(self, el, text):
        """Convert table rows.

        NB: rows are processed before the table tag is.
        Spaces to fill out columns are added in post-processing!

        Examples:
            >>> import html2md
            >>> h = '\
            <table>\
              <tr>\
                <th>th1aaa</th><th>th2</th>\
              </tr>\
              <tr>\
                <td>td1</td><td>td2</td>\
              </tr>\
            </table>'
            >>> html2md.markdownify(h)
            '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'

            i.e.:
                | th1aaa | th2 |
                | td1    | td2 |
        """
##        def wrap_cell_text(cell_text):
##            while len(cell_text) < self.max_len:
##                cell_text += "ç" # placeholder; will be replaced by spaces later
##            return cell_text
##
##        try:
##            self.max_len
##        except:
##            self.max_len = 0
##        if self.max_len == 0:
##            max_len = 0
##            for desc in el.find_all():
##                if len(desc.text) > max_len:
##                    max_len = len(desc.text)
##            if max_len>30:
##                max_len = 30
##            self.max_len = max_len

        t = []
        if el.find('th'): # headers:
            for th in el.find_all('th'):
                #t.append(wrap_cell_text(th.text))
                t.append(th.text)
            t = "|".join(t)
            return '|{}|\n|{}|\n'.format(t, self.create_underline_line(t, '-'))
        else:
            for td in el.find_all('td'):
                #t.append(wrap_cell_text(td.text))
                t.append(td.get_text(" ", strip=True))
            return '|{}|\n'.format("|".join(t))

    convert_ul = convert_list

    def convert_strong(self, el, text):
        """Convert <b> and <strong> tags.

        NB: convert_b refers to this same function

        Examples:
            >>> import html2md
            >>> h = 'abc <strong>def</strong> ghi'
            >>> html2md.markdownify(h)
            'abc **def** ghi'

            >>> import html2md
            >>> h = 'abc <b>def</b> ghi'
            >>> html2md.markdownify(h)
            'abc **def** ghi'
        """
        return '**%s**' % text if text else ''




def markdownify(html, **options):
    """Convert html to markdown.

    Calls the convert() method of the MarkdownConverter class.
    """
    return MarkdownConverter(**options).convert(html)


if __name__ == "__main__":
    test_html = """
    <?xml version="1.0" encoding="UTF-8"?>
    <html xml:lang="ar" lang="ar" dir="rtl" xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
    <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="initial-scale=2.3, user-scalable=no" />
    <title>TITLE</title>
    <link rel="stylesheet" type="text/css" href="../Style/epub.css" />
    </head>
    <body>
    <div dir="rtl" id="1">
      <h1 dir="rtl" class="title center">H1 TITLE</h1>
      <div dir="rtl" class="subtitle" id="2">SUBTITLE DIV
      </div>
      <div dir="rtl" class="section" id="sect1_34">
        <h4 dir="rtl" class="title center">H4 SECTION TITLE</h4>
        <div class="center" id="4">
          <p>P BLA <span class="quran">QURAN QUOTE</span> P BLABLA</p>
        </div>
      </div>
      <div dir="rtl" class="section" id="sect1_35">
        <h4 dir="rtl" class="title">H4 SECTION TITLE</h4>
        <p>P1

        </p>
        <p>
                            P2
                        P2b

        </p>
        <div dir="rtl" class="section" id="sect2_3">
          <h4 dir="rtl" class="title">H4 SUBSECTION2_3 TITLE</h4>
          <p>SUBSECTION2_3 PARAGRAPH</p>
        </div>
      </div>
    </div>
    </body>
    </html>
    """

    import doctest
    doctest.testmod()

    soup = BeautifulSoup(test_html)
    text_div = soup.html.body.div
    #print(text_div)
    #text=markdownify(text_div, md_style=ATX)
    #print("*"*40)
    #print(text)








import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)

from openiti.git.get_issues import get_issues


def list_issues(repo_name, access_token=None,
               #user=None, password=None,
               issue_labels=None, state="open"):
    issues = get_issues(repo_name,
                        issue_labels=issue_labels,
                        state=state,
                        access_token=access_token)
    issue_list = []
    problematic_issues = []
    for issue in issues:
        #print(issue.title)
        #print(issue.body)
        body = "\n".join(re.split("[ .]*[\r\n]+", issue.body))
        comments = issue.get_comments()
        if comments.totalCount>0:
            comments = [c.body for c in comments]
            #print(comments)
        else:
            comments = []
        
        try: 
            old_uri = re.findall("OLD URI:.*?([a-zA-Z\d.\-]+)\n", body)[0]
            new_uri = re.findall("NEW URI:.*?([a-zA-Z\d.\-]+)\n", body)[0]
            reason = re.findall("REASON:([^\n]+)", body)[0].strip()
            comments = [reason,] + comments

            #print(old_uri, issue.title, old_uri==issue.title.strip())
            #print(new_uri)
            if old_uri==issue.title.strip():
                issue_list.append([str(issue.number),
                                   old_uri,
                                   new_uri,
                                   "::".join(comments),
                                   issue.html_url
                                   ])
        except:
            #print("extraction failed")
            #input("continue?")
            problematic_issues.append(issue)
        #print("***")
    with open("issues.csv", mode="w", encoding="utf-8") as file:
        issues = "\n".join(["\t".join(i) for i in sorted(issue_list)])
        file.write(issues)
        print(issues)
    return issue_list

if __name__ == "__main__": 

    tok_fp = r"D:\London\OpenITI\metadata\automation\kitab-metadata-automation\kitab-metadata-automation\GitHub personalAccessTokenReadOnly.txt"
    with open(tok_fp, mode="r") as file:
        tok = file.read()
    issues = list_issues("OpenITI/Annotation",
                         issue_labels=["URI change suggestion"],
                         state="open", access_token=tok)

    input("Continue?")
    issues = define_text_uris(issues)
    uri_dict = sort_issues_by_uri(issues)
    print_issues_by_uri(uri_dict, "test.tsv")



"""
Test relative speed of string.replace, re.sub replacement
and mixed replacement (using re.sub for character disjunctions
like [أاإآ] and string replace for single replacements)
for the denoise and normalize functions.

Conclusion of the test:

With both tested strings (a 30-line fully vocalized string and a 75MB text),
string replacement is almost always significantly faster
than re.sub and mixed replacement;
Only when a single regex is used to replace many characters by an empty string,
a single re.sub operation with a compiled regex is rarely (!)
marginally faster than multiple string replace operations
in very large texts or after 100.000 loops in the short string.
(tested up to 25 string replace operations for one single regex)


Output: 

----------------------------------------
Testing functions re.sub denoise, string replace denoise:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - string replace denoise: 0.008388800000000002
2 - re.sub denoise: 0.0274035
(string replace denoise ca. 3.3 times faster than re.sub denoise)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - string replace denoise: 0.8215248000000002
2 - re.sub denoise: 2.6362182
(string replace denoise ca. 3.2 times faster than re.sub denoise)

----------------------------------------
Testing functions re.sub denoise, string replace denoise:

Results after 10 loops, with large text of 75MB:
1 - re.sub denoise: 4.5313052
2 - string replace denoise: 6.2427817999999995
(re.sub denoise ca. 1.4 times faster than string replace denoise)

Results after 50 loops, with large text of 75MB:
1 - re.sub denoise: 21.9464538
2 - string replace denoise: 31.245985300000008
(re.sub denoise ca. 1.4 times faster than string replace denoise)

----------------------------------------
Testing functions re.sub normalization, string replace normalization, mixed normalization:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - string replace normalization: 0.002014299999999025
2 - mixed normalization: 0.004392899999999145
3 - re.sub normalization: 0.01432450000000074
(string replace normalization ca. 2.2 times faster than mixed normalization)
(string replace normalization ca. 7.1 times faster than re.sub normalization)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - string replace normalization: 0.1923208000000045
2 - mixed normalization: 0.4523312999999973
3 - re.sub normalization: 1.357212600000011
(string replace normalization ca. 2.4 times faster than mixed normalization)
(string replace normalization ca. 7.1 times faster than re.sub normalization)

Results after 100000 loops, with 30 first lines of the Qur'an:
1 - string replace normalization: 1.9453049999999905
2 - mixed normalization: 4.5060936
3 - re.sub normalization: 18.344620699999993
(string replace normalization ca. 2.3 times faster than mixed normalization)
(string replace normalization ca. 9.4 times faster than re.sub normalization)

----------------------------------------
Testing functions re.sub normalization, string replace normalization, mixed normalization:

Results after 10 loops, with large text of 75MB:
1 - string replace normalization: 5.748965499999997
2 - mixed normalization: 11.941127800000004
3 - re.sub normalization: 34.2564823
(string replace normalization ca. 2.1 times faster than mixed normalization)
(string replace normalization ca. 6.0 times faster than re.sub normalization)

Results after 50 loops, with large text of 75MB:
1 - string replace normalization: 31.95850839999997
2 - mixed normalization: 64.00169269999998
3 - re.sub normalization: 169.95990599999996
(string replace normalization ca. 2.0 times faster than mixed normalization)
(string replace normalization ca. 5.3 times faster than re.sub normalization)

----------------------------------------
Testing functions re.sub denormalization, string replace denormalization:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - string replace denormalization: 0.009504700000000001
2 - re.sub denormalization: 0.0194251
(string replace denormalization ca. 2.0 times faster than re.sub denormalization)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - string replace denormalization: 0.9411094000000002
2 - re.sub denormalization: 1.9976751000000001
(string replace denormalization ca. 2.1 times faster than re.sub denormalization)

Results after 100000 loops, with 30 first lines of the Qur'an:
1 - string replace denormalization: 9.413773800000001
2 - re.sub denormalization: 19.1022952
(string replace denormalization ca. 2.0 times faster than re.sub denormalization)

----------------------------------------
Testing functions re.sub denormalization, string replace denormalization:

Results after 10 loops, with large text of 75MB:
1 - string replace denormalization: 25.790224800000004
2 - re.sub denormalization: 43.2591982
(string replace denormalization ca. 1.7 times faster than re.sub denormalization)

Results after 50 loops, with large text of 75MB:
1 - string replace denormalization: 135.18536029999996
2 - re.sub denormalization: 222.3127232
(string replace denormalization ca. 1.6 times faster than re.sub denormalization)

----------------------------------------
Testing functions single re.sub operation, 5 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 5 string replace operations: 0.0032636999999999996
2 - single re.sub operation: 0.006046800000000001
(5 string replace operations ca. 1.9 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 5 string replace operations: 0.3252406000000001
2 - single re.sub operation: 0.5332258
(5 string replace operations ca. 1.6 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 5 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 5 string replace operations: 7.272236500000002
2 - single re.sub operation: 17.459449199999998
(5 string replace operations ca. 2.4 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 6 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 6 string replace operations: 0.003712800000002403
2 - single re.sub operation: 0.005911600000001016
(6 string replace operations ca. 1.6 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 6 string replace operations: 0.37006929999999727
2 - single re.sub operation: 0.5517581000000007
(6 string replace operations ca. 1.5 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 6 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 6 string replace operations: 8.367578600000002
2 - single re.sub operation: 16.8720289
(6 string replace operations ca. 2.0 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 7 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 7 string replace operations: 0.0044642000000010285
2 - single re.sub operation: 0.006948700000002361
(7 string replace operations ca. 1.6 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 7 string replace operations: 0.43168089999999637
2 - single re.sub operation: 0.6184892000000062
(7 string replace operations ca. 1.4 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 7 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 7 string replace operations: 9.770164199999996
2 - single re.sub operation: 19.609205799999998
(7 string replace operations ca. 2.0 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 8 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 8 string replace operations: 0.005006100000002789
2 - single re.sub operation: 0.00739210000000412
(8 string replace operations ca. 1.5 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 8 string replace operations: 0.495628300000007
2 - single re.sub operation: 0.677892600000007
(8 string replace operations ca. 1.4 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 8 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 8 string replace operations: 12.027051700000001
2 - single re.sub operation: 23.323600100000007
(8 string replace operations ca. 1.9 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 9 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 9 string replace operations: 0.0055399000000022625
2 - single re.sub operation: 0.007424799999995457
(9 string replace operations ca. 1.3 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 9 string replace operations: 0.5545536999999996
2 - single re.sub operation: 0.7581914000000012
(9 string replace operations ca. 1.4 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 9 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 9 string replace operations: 12.158491499999997
2 - single re.sub operation: 25.170432899999994
(9 string replace operations ca. 2.1 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 10 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 10 string replace operations: 0.006212199999993118
2 - single re.sub operation: 0.007610900000003085
(10 string replace operations ca. 1.2 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 10 string replace operations: 0.6160366999999951
2 - single re.sub operation: 0.7125027999999816
(10 string replace operations ca. 1.2 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 10 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 10 string replace operations: 14.283474799999993
2 - single re.sub operation: 23.337247899999994
(10 string replace operations ca. 1.6 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 11 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 11 string replace operations: 0.0067801000000144995
2 - single re.sub operation: 0.007790099999994027
(11 string replace operations ca. 1.1 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 11 string replace operations: 0.663314999999983
2 - single re.sub operation: 0.8159454000000039
(11 string replace operations ca. 1.2 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 11 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 11 string replace operations: 14.236086099999994
2 - single re.sub operation: 29.64436070000002
(11 string replace operations ca. 2.1 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 12 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 12 string replace operations: 0.007327099999997699
2 - single re.sub operation: 0.00877140000000054
(12 string replace operations ca. 1.2 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - single re.sub operation: 0.8271442999999863
2 - 12 string replace operations: 0.9188682999999855
(single re.sub operation ca. 1.1 times faster than 12 string replace operations)

----------------------------------------
Testing functions single re.sub operation, 12 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 12 string replace operations: 20.162684599999977
2 - single re.sub operation: 31.15590209999999
(12 string replace operations ca. 1.5 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 13 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 13 string replace operations: 0.008608600000002298
2 - single re.sub operation: 0.012064600000030623
(13 string replace operations ca. 1.4 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - 13 string replace operations: 1.1774655000000394
2 - single re.sub operation: 1.7925384000000122
(13 string replace operations ca. 1.5 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 13 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 13 string replace operations: 17.298843399999953
2 - single re.sub operation: 36.47844249999997
(13 string replace operations ca. 2.1 times faster than single re.sub operation)

----------------------------------------
Testing functions single re.sub operation, 14 string replace operations:

Results after 100 loops, with 30 first lines of the Qur'an:
1 - 14 string replace operations: 0.009635100000025432
2 - single re.sub operation: 0.014636200000040844
(14 string replace operations ca. 1.5 times faster than single re.sub operation)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - single re.sub operation: 1.5175319999999601
2 - 14 string replace operations: 1.7029174000000467
(single re.sub operation ca. 1.1 times faster than 14 string replace operations)

----------------------------------------
Testing functions single re.sub operation, 14 string replace operations:

Results after 10 loops, with large text of 75MB:
1 - 14 string replace operations: 19.107556999999986
2 - single re.sub operation: 43.3557414
(14 string replace operations ca. 2.3 times faster than single re.sub operation)

"""

import re
import timeit


def time_functions(functions, loops, string_descr):
    """Check which function is faster.

    Args:
        functions (list): a list of tuples (function_name (str),
                                            function_descr (str))
        loops (list): a list of integers, representing the numbers
            of times a pair of functions should be timed.
        string_descr (str): a description of the test string.
    NB: the functions to be tested should not have arguments.
    """
    print("--"*20)
    print("Testing functions {}:".format(", ".join([y for x,y in functions])))
    print()
    accuracy = (len(set([eval(f[0]+'()') for f in functions])) == 1)
    print("All functions have the same outcome: ", accuracy)
    for n in loops:
        results = []
        for f, f_descr in functions:
            t = timeit.timeit(f+"()", setup="from __main__ import "+f, number=n)
            results.append([f_descr, t])
        results = sorted(results, key=lambda item: item[1])
        print("Results after {} loops, with {}:".format(n, string_descr))
        for i, x in enumerate(results):
            print("{} - {}: {}".format(i+1, x[0], x[1]))
        for x in results[1:]:
            fmt = "({} ca. {} times faster than {})"
            print(fmt.format(results[0][0], round(x[1]/results[0][1], 1), x[0]))
        print()



def re_sub_deNoise(text):
    noise = re.compile(""" ّ    | # Tashdid
                           َ    | # Fatha
                           ً    | # Tanwin Fath
                           ُ    | # Damma
                           ٌ    | # Tanwin Damm
                           ِ    | # Kasra
                           ٍ    | # Tanwin Kasr
                           ْ    | # Sukun
                           ـ     # Tatwil/Kashida
                         """, re.VERBOSE)
    return re.sub(noise, '', text)


def repl_deNoise(text):
    text = text.replace("ّ", "")  # Tashdid
    text = text.replace("َ", "")  # Fatha
    text = text.replace("ً", "")  # Tanwin Fath
    text = text.replace("ُ", "")  # Damma
    text = text.replace("ٌ", "")  # Tanwin Damm
    text = text.replace("ِ", "")  # Kasra
    text = text.replace("ٍ", "")  # Tanwin Kasr
    text = text.replace("ْ", "")  # Sukun
    text = text.replace("ـ", "") # Tatwil/Kashida
    return text

def re_sub_deNoise_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return re_sub_deNoise(text)

def repl_deNoise_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return repl_deNoise(text)



def re_sub_normalizeArabic(text):
    text = re.sub("[إأٱآ]", "ا", text)
    text = re.sub("ى", "ي", text)
    text = re.sub("(ؤ)", "و", text)
    text = re.sub("(ئ)", "ي", text)
    text = re.sub("(ة)", "ه", text)
    return text

def repl_normalizeArabic(text):
    text = text.replace("أ", "ا")
    text = text.replace("ٱ", "ا")
    text = text.replace("آ", "ا")
    text = text.replace("إ", "ا")
    text = text.replace("ى", "ي")
    text = text.replace("ؤ", "و")
    text = text.replace("ئ", "ي")
    text = text.replace("ة", "ه")
    return text


def mix_normalizeArabic(text):
    text = re.sub("[إأٱآ]", "ا", text)
    text = text.replace("ى", "ي")
    text = text.replace("ؤ", "و")
    text = text.replace("ئ", "ي")
    text = text.replace("ة", "ه")
    return text


def re_sub_normalizeArabic_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return re_sub_normalizeArabic(text)

def repl_normalizeArabic_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return repl_normalizeArabic(text)

def mix_normalizeArabic_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return mix_normalizeArabic(text)


def re_sub_denormalize(text):
    """Replace complex characters with a regex covering all variants."""
    alifs = '[إأٱآا]'
    alif_maqsura = '[يى]\b'
    ta_marbutas = '[هة]\b'
    hamzas = '[ؤئء]'
    hamzas_reg = '[ؤئءوي]'
    # Applying deNormalization
    text = re.sub(alifs, alifs, text)
    text = re.sub(alif_maqsura, alif_maqsura, text)
    text = re.sub(ta_marbutas, ta_marbutas, text)
    text = re.sub(hamzas, hamzas_reg, text)
    return text

def repl_denormalize(text):
    """Replace complex characters with a regex covering all variants."""
    text = text.replace("أ", "[إأٱآا]")
    text = text.replace("ٱ", "[إأٱآا]")
    text = text.replace("آ", "[إأٱآا]")
    text = text.replace("إ", "[إأٱآا]")
    text = text.replace("ى", "[يى]")
    text = text.replace("ي", "[يى]")
    text = text.replace("ؤ", "[ؤئءوي]")
    text = text.replace("ئ", "[ؤئءوي]")
    text = text.replace("ء", "[ؤئءوي]")
    text = text.replace("ة", "[هة]")
    return text

def re_sub_denormalize_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return re_sub_denormalize(text)

def repl_denormalize_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return repl_denormalize(text)


def single_re_sub_empty(text, repl_chars):
    """Remove specific characters from a string using a single regex."""
    repl = re.compile("[{}]".format(repl_chars))
    return repl.sub("", text)

def multiple_str_repl_empty(text, repl_chars):
    """Remove specific characters from a string
    using multiple string replace operations."""
    for c in repl_chars:
        text = text.replace(c, "")
    return text

def single_re_sub_empty_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return single_re_sub_empty(text, repl_chars)

def multiple_str_repl_empty_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return multiple_str_repl_empty(text, repl_chars)



def single_re_sub(text, repl_chars):
    """Replace specific characters with another character using a single regex."""
    repl = re.compile("[{}]".format(repl_chars))
    return repl.sub("a", text)

def multiple_str_repl(text, repl_chars):
    """Replace specific characters with another character
    using multiple string replace operations."""
    for c in repl_chars:
        text = text.replace(c, "a")
    return text

def single_re_sub_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return single_re_sub(text, repl_chars)

def multiple_str_repl_test():
    """A function without arguments, needed for timeit.
    NB: uses global variable text to solve the argument problem"""
    return multiple_str_repl(text, repl_chars)


page = """
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ
الْحَمْدُ لِلَّهِ رَبِّ الْعَالَمِينَ
الرَّحْمَنِ الرَّحِيمِ
مَالِكِ يَوْمِ الدِّينِ
إِيَّاكَ نَعْبُدُ وَإِيَّاكَ نَسْتَعِينُ
اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ
صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ الم
ذَلِكَ الْكِتَابُ لَا رَيْبَ فِيهِ هُدًى لِلْمُتَّقِينَ
الَّذِينَ يُؤْمِنُونَ بِالْغَيْبِ وَيُقِيمُونَ الصَّلَاةَ وَمِمَّا رَزَقْنَاهُمْ يُنْفِقُونَ
وَالَّذِينَ يُؤْمِنُونَ بِمَا أُنْزِلَ إِلَيْكَ وَمَا أُنْزِلَ مِنْ قَبْلِكَ وَبِالْآخِرَةِ هُمْ يُوقِنُونَ
أُولَئِكَ عَلَى هُدًى مِنْ رَبِّهِمْ وَأُولَئِكَ هُمُ الْمُفْلِحُونَ
إِنَّ الَّذِينَ كَفَرُوا سَوَاءٌ عَلَيْهِمْ أَأَنْذَرْتَهُمْ أَمْ لَمْ تُنْذِرْهُمْ لَا يُؤْمِنُونَ
خَتَمَ اللَّهُ عَلَى قُلُوبِهِمْ وَعَلَى سَمْعِهِمْ وَعَلَى أَبْصَارِهِمْ غِشَاوَةٌ وَلَهُمْ عَذَابٌ عَظِيمٌ
وَمِنَ النَّاسِ مَنْ يَقُولُ آمَنَّا بِاللَّهِ وَبِالْيَوْمِ الْآخِرِ وَمَا هُمْ بِمُؤْمِنِينَ
يُخَادِعُونَ اللَّهَ وَالَّذِينَ آمَنُوا وَمَا يَخْدَعُونَ إِلَّا أَنْفُسَهُمْ وَمَا يَشْعُرُونَ
فِي قُلُوبِهِمْ مَرَضٌ فَزَادَهُمُ اللَّهُ مَرَضًا وَلَهُمْ عَذَابٌ أَلِيمٌ بِمَا كَانُوا يَكْذِبُونَ
وَإِذَا قِيلَ لَهُمْ لَا تُفْسِدُوا فِي الْأَرْضِ قَالُوا إِنَّمَا نَحْنُ مُصْلِحُونَ
أَلَا إِنَّهُمْ هُمُ الْمُفْسِدُونَ وَلَكِنْ لَا يَشْعُرُونَ
وَإِذَا قِيلَ لَهُمْ آمِنُوا كَمَا آمَنَ النَّاسُ قَالُوا أَنُؤْمِنُ كَمَا آمَنَ السُّفَهَاءُ أَلَا إِنَّهُمْ هُمُ السُّفَهَاءُ وَلَكِنْ لَا يَعْلَمُونَ
وَإِذَا لَقُوا الَّذِينَ آمَنُوا قَالُوا آمَنَّا وَإِذَا خَلَوْا إِلَى شَيَاطِينِهِمْ قَالُوا إِنَّا مَعَكُمْ إِنَّمَا نَحْنُ مُسْتَهْزِئُونَ
اللَّهُ يَسْتَهْزِئُ بِهِمْ وَيَمُدُّهُمْ فِي طُغْيَانِهِمْ يَعْمَهُونَ
أُولَئِكَ الَّذِينَ اشْتَرَوُا الضَّلَالَةَ بِالْهُدَى فَمَا رَبِحَتْ تِجَارَتُهُمْ وَمَا كَانُوا مُهْتَدِينَ
مَثَلُهُمْ كَمَثَلِ الَّذِي اسْتَوْقَدَ نَارًا فَلَمَّا أَضَاءَتْ مَا حَوْلَهُ ذَهَبَ اللَّهُ بِنُورِهِمْ وَتَرَكَهُمْ فِي ظُلُمَاتٍ لَا يُبْصِرُونَ
صُمٌّ بُكْمٌ عُمْيٌ فَهُمْ لَا يَرْجِعُونَ
أَوْ كَصَيِّبٍ مِنَ السَّمَاءِ فِيهِ ظُلُمَاتٌ وَرَعْدٌ وَبَرْقٌ يَجْعَلُونَ أَصَابِعَهُمْ فِي آذَانِهِمْ مِنَ الصَّوَاعِقِ حَذَرَ الْمَوْتِ وَاللَّهُ مُحِيطٌ بِالْكَافِرِينَ
يَكَادُ الْبَرْقُ يَخْطَفُ أَبْصَارَهُمْ كُلَّمَا أَضَاءَ لَهُمْ مَشَوْا فِيهِ وَإِذَا أَظْلَمَ عَلَيْهِمْ قَامُوا وَلَوْ شَاءَ اللَّهُ لَذَهَبَ بِسَمْعِهِمْ وَأَبْصَارِهِمْ إِنَّ اللَّهَ عَلَى كُلِّ شَيْءٍ قَدِيرٌ
يَا أَيُّهَا النَّاسُ اعْبُدُوا رَبَّكُمُ الَّذِي خَلَقَكُمْ وَالَّذِينَ مِنْ قَبْلِكُمْ لَعَلَّكُمْ تَتَّقُونَ
الَّذِي جَعَلَ لَكُمُ الْأَرْضَ فِرَاشًا وَالسَّمَاءَ بِنَاءً وَأَنْزَلَ مِنَ السَّمَاءِ مَاءً فَأَخْرَجَ بِهِ مِنَ الثَّمَرَاتِ رِزْقًا لَكُمْ فَلَا تَجْعَلُوا لِلَّهِ أَنْدَادًا وَأَنْتُمْ تَعْلَمُونَ
وَإِنْ كُنْتُمْ فِي رَيْبٍ مِمَّا نَزَّلْنَا عَلَى عَبْدِنَا فَأْتُوا بِسُورَةٍ مِنْ مِثْلِهِ وَادْعُوا شُهَدَاءَكُمْ مِنْ دُونِ اللَّهِ إِنْ كُنْتُمْ صَادِقِينَ
"""

fp = r"D:\London\OpenITI\bok\ShamelaRARs\converted\4_OpenITI_mARkdown_new\Shamela026332.txt"
with open(fp, mode="r", encoding="utf-8") as file:
    large_text = file.read()

##    
##text = page  # set the global text variable
##
##time_functions([("re_sub_deNoise_test", "re.sub denoise"),
##                ("repl_deNoise_test", "string replace denoise"),],
##               [100, 10000],
##               "30 first lines of the Qur'an")
##
##text = large_text  # reset the global text variable
##
##time_functions([("re_sub_deNoise_test", "re.sub denoise"),
##                ("repl_deNoise_test", "string replace denoise")],
##               [10, 50],
##               "large text of 75MB")
##
##
##text = page  # reset the global text variable
##
##time_functions([("re_sub_normalizeArabic_test", "re.sub normalization"),
##                ("repl_normalizeArabic_test", "string replace normalization"),
##                ("mix_normalizeArabic_test", "mixed normalization")],
##               [100, 10000, 100000],
##               "30 first lines of the Qur'an")
##
##text = large_text  # reset the global text variable
##
##time_functions([("re_sub_normalizeArabic_test", "re.sub normalization"),
##                ("repl_normalizeArabic_test", "string replace normalization"),
##                ("mix_normalizeArabic_test", "mixed normalization")],
##               [10, 50],
##               "large text of 75MB")

##text = page  # reset the global text variable
##
##time_functions([("re_sub_denormalize_test", "re.sub denormalization"),
##                ("repl_denormalize_test", "string replace denormalization")],
##               [100, 10000, 100000],
##               "30 first lines of the Qur'an")
##
##text = large_text  # reset the global text variable
##
##time_functions([("re_sub_denormalize_test", "re.sub denormalization"),
##                ("repl_denormalize_test", "string replace denormalization")],
##               [10, 50],
##               "large text of 75MB")

##chars = "جحخهعغفقثصضكمنتالبيسشزوةىرؤءئ"
##for x in range(15, 25):
##    repl_chars = chars[:x]
##    text = page  # reset the global text variable
##
##    time_functions([("single_re_sub_test", "single re.sub operation"),
##                    ("multiple_str_repl_test",
##                     "{} string replace operations".format(len(repl_chars)))],
##                   [100, 10000],
##                   "30 first lines of the Qur'an")
##
##    text = large_text  # reset the global text variable
##
##    time_functions([("single_re_sub_test", "single re.sub operation"),
##                    ("multiple_str_repl_test",
##                     "{} string replace operations".format(len(repl_chars)))],
##                    [10],
##                   "large text of 75MB")

chars = "جحخهعغفقثصضكمنتالبيسشزوةىرؤءئ"
for x in range(5, 25):
    repl_chars = chars[:x]
    text = page  # reset the global text variable

    time_functions([("single_re_sub_empty_test", "single re.sub remove"),
                    ("multiple_str_repl_empty_test",
                     "{} string replace operations".format(len(repl_chars)))],
                   [100, 10000, 100000],
                   "30 first lines of the Qur'an")

    text = large_text  # reset the global text variable

    time_functions([("single_re_sub_empty_test", "single re.sub remove"),
                    ("multiple_str_repl_empty_test",
                     "{} string replace operations".format(len(repl_chars)))],
                    [10, 50],
                   "large text of 75MB")




import os
from collections import Counter
import re


def collect_book_ids(source_dir):
    """
    Collects book ids of the OpenITI corpus and return a dictionary of source identifiers
    and the, frequencies. Source identifier is the first part of a book id that specifies
    the source form which the book comes, e.g., Shamela, JK, etc.

    Output sample:
    {'Shamela': 3691, 'JK': 2309}

    Args:
        source_dir: path to the folder containing OpenITI texts or XXXXAH repos

    Returns:
        Dictionary: source ids to frequency

    """

    id_set = set()
    for root, dirs, files in os.walk(source_dir):
        for f in files:
            txt_f = re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))?$", f)
            if txt_f:
                b_id = re.sub("-\w{4}", "", txt_f.group(0).split(".")[2])
                id_set.add(b_id)

    lib_ids = [re.sub("\d+(Vols)?([A-Z]{1})?(BK\d{1})?$", "", x) for x in id_set]
    return Counter(lib_ids)


if __name__ == '__main__':

    source = input("Enter the source directory of all AH repositories: ")

    if not os.path.exists(source):
        print("No such directory %s " % source)
    else:
        collect_book_ids(source)



# Copies all data from the XXXXAH directories to a single directory in order to publish a version of OpenITI.

import os
import shutil
import sys
import re


def copy_ah_data(source_dir, dest_dir):
    # All XXXXAH direcortories
    ah_dirs = os.listdir(source_dir)
    for ad in ah_dirs:
        # Get the path to the "data" folder (in each XXXXAH dir)
        data_path = os.path.join(source_dir, ad, "data")
        # If the data directory exists (for cases the "data" dir is not available, like 1475AH)
        if os.path.exists(data_path):
            # Get the list of directories in "data" directory
            cur_dirs = os.listdir(data_path)
            # Copy the dirs in "/data"
            for cpy_d in cur_dirs:
                # Source is the path to the current folder in data folder (cpy_d) that will be copied,
                # Target is a join of the target path (given as input) and the current folder (cpy_d)
                shutil.copytree(os.path.join(data_path, cpy_d), os.path.join(dest_dir, cpy_d))
                for root, dirs, files in os.walk(os.path.join(dest_dir, cpy_d)):
                # texts = [f for f in files if
                #          os.path.isfile(os.path.join(dest_dir, cpy_d, f)) and
                #          re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))?$", f)]
                # texts_noExt = set([re.split("\.(mARkdown|inProgress|completed)", t)[0] for t in texts])
                    for f in files:
                        no_ext_file = re.split("\.(mARkdown|inProgress|completed)", f)[0]
                        no_ext_path = os.path.join(root, no_ext_file)
                        if f.endswith(".mARkdown"):
                            if os.path.exists(no_ext_path + ".completed"):
                            #     try:
                                os.remove(no_ext_path + ".completed")
                                # except OSError:
                                #     pass
                            if os.path.exists(no_ext_path + ".inProgress"):
                                #     try:
                                os.remove(no_ext_path + ".inProgress")
                                    # except OSError:
                                    #     pass
                        elif f.endswith(".completed"):
                            #     try:
                            if os.path.exists(no_ext_path + ".inProgress"):
                                os.remove(no_ext_path + ".inProgress")
                                # except OSError:
                                #     pass

                        elif f.endswith(".inProgress"):
                            # try:
                            os.remove(os.path.join(root, f))
                            # except OSError:
                            #     pass

        else:
            print("%s repository doesn't have any 'data' directory!" % ad)


if __name__ == '__main__':

    source = input("Enter the source directory: ")
    target = input("Enter the target directory: ")

    if len(sys.argv) > 0:
        if not os.path.exists(source):
            print("source directory doesn't exists. Re-run the script and give the source!")
        else:
            copy_ah_data(source, target)
    else:
        print("Give the path to the script...!")




"""A generic converter for converting tei xml files into OpenITI format.

Subclass the TeiConverter to make a converter for a new
source of scraped books in TEI xml format.

Examples:
    Generic tei conversion::
    
        >>> from tei_converter_generic import TeiConverter
        >>> conv = TeiConverter(dest_folder="test/converted")
        >>> conv.VERBOSE = False
        >>> folder = r"test"
        >>> fn = r"GRAR000070"
        >>> conv.convert_file(os.path.join(folder, fn))
        >>> conv.convert_files_in_folder(folder, extensions = [""])

    Sub-class to create a converter specific to GRAR collection tei files::
    
        #>>> class GRARConverter(TeiConverter):
        #        def post_process(self, text):
        #            text = super().post_process(text)
        #            vols = re.findall("#META#.+ vol. (\d+) pp", text)
        #            if len(vols) == 1:
        #                text = re.sub("PageV00", "PageV{:02d}".format(int(vols[0])), text)
        #            elif len(vols) == 0:
        #                text = re.sub("PageV00", "PageV01", text)
        #            return text
        #>>> conv = GRARConverter("test/converted")
        #>>> conv.VERBOSE = False
        #>>> folder = r"test"
        #>>> conv.convert_files_in_folder(folder, exclude_ext = ["py", "yml", "txt", "epub"])

================================== ==========================
GenericConverter                   TeiConverter
================================== ==========================
__init__                           __init__ (appended)
convert_files_in_folder            (inherited)
convert file                       (inherited)
make_dest_fp                       (inherited)
get_metadata (dummy)               get_metadata
get_data                           (inherited)
pre_process                        pre_process
add_page_numbers (dummy)           (inherited - not used)
add_structural_annotations (dummy) add_structural_annotations
remove_notes (dummy)               (inherited - generic!)
reflow                             (inherited)
add_milestones (dummy)             (inherited - dummy)
post_process                       post_process
compose                            (inherited)
save_file                          (inherited)
                                   preprocess_page_numbers
                                   preprocess_wrapped_lines
================================== ==========================
"""

from bs4 import BeautifulSoup
import os
import re
import textwrap

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.helper.ara import deNoise
from openiti.new_books.convert.generic_converter import GenericConverter
from openiti.new_books.convert.helper import tei2md

class TeiConverter(GenericConverter):

    def __init__(self, dest_folder=None):
        """Initialize the class and its super-class."""
        super().__init__()
        self.root_tag_string = "soup.html.body"
##        self.unique_tags = dict()



    def pre_process(self, text):
        """Remove unwanted features from the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its initial state

        Returns:
            text (str): pre-processed text
        """
        super().pre_process(text)
        text = self.preprocess_page_numbers(text)
        text = self.preprocess_wrapped_lines(text)
        return text


    def preprocess_page_numbers(self, s):
        """Turn page beginnings into page endings.

        TEI xml indicates the beginning of a page, while OpenITI mARkdown
        indicates the end of a page.
        """
        try:
            s, tail = re.split("</body>", s, 1)
        except:
            tail = ""
        spl = re.split('<pb n="(\d+)"/?>', s)
        if len(spl) < 3:
            spl = re.split('<span [^>]+pb-(\d+)[^>]+>', s)
        s = spl
        page_numbers = [it for it in s if s.index(it)%2]
        page_texts = [it for it in s if not s.index(it)%2]
        if page_texts[0] != "":
            page_numbers = ["0"] + page_numbers
        else:
            page_texts = page_texts[1:]
        text = ""
        for i, it in enumerate(page_texts):
            text += it
            text += "\n\nPageV00P{:03d}\n".format(int(page_numbers[i]))
        if tail:
            return text + "</body>" + tail
        else:
            return text

    def preprocess_wrapped_lines(self, s):
        s = re.sub("-[\n\r]+~~", "", s)
        s = re.sub("[\n\r]+~~", " ", s)
        return re.sub(" {2,}", " ", s)


    def get_metadata(self, text):
        """Extracts the metadata from the text.


        Args:
            text (str): text of the file that contains data + metadata

        Returns:
            metadata (str): metadata formatted in OpenITI format
                (including the magic value and the header splitter)
        """
        soup = BeautifulSoup(text, "xml")
        metadata = ""
        header = soup.find("teiHeader")
        file_d = ["author", "title"]
        fileDesc = header.find("fileDesc")
        for k in file_d:
            try:
                item = fileDesc.find(k).text.strip()
            except:
                item = ""
            metadata += "#META# {}: {}\n".format(k, item)

        source_d = {"title": "", "author": "n.n.", "editor": "n.n.",
                    "pubPlace": "n.p.", "publisher": "n.n.", "date": "n.d."}
        sourceDesc = header.find("sourceDesc")
        for k in source_d:
            try:
                source_d[k] = sourceDesc.find(k).text.strip()
            except:
                pass
        p = ""
        page_info = sourceDesc.find_all("biblScope")
        for i in page_info:
            if i.has_attr("unit"):
                if i["unit"] == "vol":
                    p += "vol. {} ".format(i.text)
                if i["unit"] in ["pp", "pages"]:
                    p += "pp. {}".format(i.text)
        ed_info = "Ed. {editor} ({date}), {title}, {pubPlace}: {publisher}"
        ed_info = ed_info.format(**source_d)
        if p:
            ed_info += ", {}.".format(p)
        else:
            ed_info += "."
        metadata += "#META# ed_info: {}\n".format(ed_info)

        text_tag = soup.find("text")
        text_id = text_tag["id"]
        metadata += "#META# coll_id: {}\n".format(text_id)
        metadata = self.magic_value + metadata + self.header_splitter
        return metadata


    def add_structural_annotations(self, text, options=None):
        soup = BeautifulSoup(text, "xml")
        text_soup = soup.find("text")
        text = tei2md.markdownify(str(text_soup), options=options)
        return text


    def post_process(self, text):
        """Carry out post-processing operations on the text.

        Args:
            text (str): the text in its current conversion state

        Returns:
            processed (str): the text after post-processing operations.
        """
        text = re.sub(r"\APageV\d+P000\n*", "", text.strip())
        processed = re.sub("[\n\r]{2,}(?![#\|P])", "\n\n# ", text)
        return processed



if __name__ == "__main__":
    import doctest
    doctest.testmod()

    input("Passed tests. Press Enter to start converting")

    conv = TeiConverter()
    #conv.convert_file("test/test.txt")
    folder = r"G:\London\OpenITI\RAWrabica\RAWrabica005000\GRAR"
    conv.convert_files_in_folder(folder, exclude_ext=["yml", "py"])
##    input()
##    source_fp = r"G:\London\OpenITI\RAWrabica\RAWrabica005000\GRAR\GRAR000070"
##    conv.convert_file(source_fp)







"""A Converter for al-Maktaba al-Shamela databases to OpenITI mARkdown format.

Al-Maktaba al-Shamela is a digital text collection, but it also 
developed its own reader, which is based on an internal database.

A large number of texts are available on its website (www.shamela.ws), 
and can be downloaded in EPUB or .bok format (the latter is a zipped .mdb file). 
These files must be converted with another script. 
See https://github.com/OpenITI/raw_SHAM19Y/wiki for that conversion process.

However, the program also allows users to add texts from other sources
(incl. their own transcriptions) to the program's database.
The database can be zipped, together with an executable for the reader,
and disseminated. Some users have used this ability
of the program to create their own collections that include many
texts not available in the original program
(for ideological or other reasons), e.g.,

* al-maktaba al-shamela al-dhahabiyya
* al-maktaba al-shamela al-ibadiyya
* al-maktaba al-shamela al-zaydiyya
* etc.

Database structure
==================

The database consists of a large number of folders and files:

* ``./Files/Main.mdb``: contains the metadata for every book in the program,
    in a table called “0bok”
* ``./Files/special.mdb``: contains more metadata + data that could be
    marked up in the text (like Qur’anic passages)
* ``./Books``: contains 10 folders, named "0" to "9", that each contain
    a large number of .mdb files (one .mdb file per text; filename = id number)

NB: in some versions of the database, the books are actually in
in large multi-book .mdb files in another folder (``./Books/Archive``),
which contains large .mdb files, each ca. 600 MB;
each of these contain the bxxxx and txxxx tables for a couple
of thousand books (xxxx being the Shamela id of the book):

- bxxxx contains the text
- txxxx contains the section titles

Module description
==================

This module contains two major components:

* ``extract_metadata``: extracts the metadata from the shamela database
* ``BokJsonConverter`` class: extracts the text and book structure
  from the books' .mdb files into json format,
  fuses this with the book's metadata,
  and converts it into OpenITI mARkdown format

The main component of the module, the BokJsonConverter class,
is a subclass of the GenericConverter from the generic_converter module:

GenericConverter
    \_ GenericEpubConverter
    \_ BokJsonConverter

Methods of both classes:
(methods of GenericConverter are inherited by GenericHtmlConverter;
methods of GenericConverter with the same name
in GenericHtmlConverter are overwritten by the latter)

=================================== =====================================
GenericConverter                    BokJsonConverter
=================================== =====================================
__init__                            __init__ (appended)
convert_files_in_folder             convert_files_in_folder
convert file                        convert file
make_dest_fp                        (inherited - generic!)
get_metadata (dummy)                (metadata collected in get_data step)
get_data                            get_data
pre_process                         pre_process
add_page_numbers (dummy)            (pages added in get_data step)
add_structural_annotations (dummy)  add_structural_annotations
remove_notes (dummy)                remove_notes
reflow                              (inherited)
add_milestones (dummy)              (inherited - dummy!)
post_process                        post_process
compose                             (inherited)
save_file                           (inherited)
                                    mdb2json
                                    format_text
                                    format_metadata
=================================== =====================================

"""


from collections import defaultdict, OrderedDict
from datetime import datetime
import json
import os
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert import generic_converter
from openiti.helper.ara import deNoise
from openiti.new_books.convert.helper import bok

VERBOSE = False



def main(meta_folder=None, files_folder=None, books_folder=None,
         conv_folder=None, download_date="", download_source="",
         config_fp=None, extensions=["mdb"], fn_regex=None, verbose=False,
         multiple_books_in_mdb=False, overwrite=True):
    """Collect the variables and carry out the conversion.

    Args:
        meta_folder (str): path to the folder containing metadata extracted
            from the database
        files_folder (str): path to the `Files` folder of the database
            (containing the main metadata database files)
        books_folder (str): path to the `Bookss` folder of the database
            (containing the .mdb files containing the text of the books)
        conv_folder (str): path to the folder where the converted files should
            be stored
        download_date (str): (approximate) date when the database was downloaded
        download_source (str): name and/or dowload url of the database
        config_fp (str): path to a config.py file containing
            the above-mentioned parameters
            NB: to print an empty config file: print_default_config_file()
    """
    # 0- import configuration if configuration file is given:

    if config_fp:
        import shutil
        shutil.copyfile(config_fp, "./temp_config.py")
        from temp_config import files_folder, books_folder, meta_folder,\
            conv_folder, download_date, download_source, fn_regex, extensions
        os.remove("temp_config.py")

    # 1a- collect metadata (extract it if it has not been extracted yet):

    if not meta_folder:
        if files_folder:
            meta_folder = os.path.join(os.path.dirname(files_folder),
                                       "extracted_metadata")
            extract_metadata(files_folder, meta_folder)
        else:
            print("Has the metadata already been extracted?")
            resp = input("Y/N? ")
            if resp.upper() == "Y":
                print("Provide the path to the folder that contains the \
metadata json files:")
                meta_folder = input()
            else:
                print("Provide the path to the `Files` folder of the \
Shamela database:")
                files_folder = input()
                if not re.match("(?:\A|[\\/])Files[\\/]?", folder):
                    msg = "folder path should end with `Files` or `Files/`"
                    raise Exception(msg)
                meta_folder = os.path.join(os.path.dirname(files_folder),
                                           "extracted_metadata")
                extract_metadata(files_folder, meta_folder)
        if not books_folder:
                  books_folder = os.path.join(os.path.dirname(meta_folder),
                                   "Books")

    if not books_folder:
        print("Provide the path to the `Books` folder of the Shamela database:")
        books_folder = input()
    if not conv_folder:
        conv_folder = books_folder+"Converted"

    meta_fp = os.path.join(meta_folder, r"main_metadata_reform.json")
    with open(meta_fp, mode="r", encoding="utf-8") as file:
        all_meta = json.load(file)

    # 1b - Add some additional metadata:

    if not download_date:
        print("Write the (approximate) date when the library was downloaded: ")
        download_date = input()
    if not download_source:
        print("write the name and/or url of the downloaded library: ")
        download_source = input()

    additional_meta = OrderedDict()
    additional_meta["DownloadSource"] = download_source
    additional_meta["DownloadDate"] = download_date
    additional_meta["ConversionDate"] = datetime.today().strftime('%Y-%m-%d')

    # 2- convert files:

    conv = BokJsonConverter(all_meta=all_meta,
                            additional_meta=additional_meta,
                            dest_folder=conv_folder,
                            multiple_books_in_mdb=multiple_books_in_mdb,
                            overwrite=overwrite)
    if verbose:
        conv.VERBOSE = True
    conv.convert_files_in_folder(books_folder,
                                 dest_folder=conv_folder, extensions=extensions,
                                 fn_regex=fn_regex)
    print("Converted files can be found in", conv_folder)
    print("(metadata in {})".format(meta_folder))


################################################################################




class BokJsonConverter(generic_converter.GenericConverter):
    """Extract book data from .mdb files into json, and convert into mARkdown.
    """

    def __init__(self, all_meta=None, additional_meta=None,
                 dest_folder=None, multiple_books_in_mdb=False,
                 overwrite=True):
        """Set up the main variables of the class.

        Args:
            all_meta (dict): dictionary representation of the
                main metadata table (./Files/Main.mdb) of the Shamela
                desktop program.
                Should only be provided for data directly derived
                from the Shamela desktop program.
                Bok-file derived jsons contain the metadata themselves.
            additional_meta (dict): a dictionary containing metadata
                not contained in the Shamela metadata
                that should be added to the metadata
        """
        super().__init__()
        self.all_meta = all_meta
        self.additional_meta = additional_meta
        if dest_folder:
            self.dest_folder = dest_folder  # otherwise: default "converted"
        self.footnote_regex = r"[\r\¶\n ]*¬?_{5,}[\r\n\¶ ]*(.*)"
        self.multiple_books_in_mdb = multiple_books_in_mdb
        self.OVERWRITE = overwrite


    def convert_file(self, source_fp, dest_fp=None):
        """Convert one file to OpenITI format.

        Args:
            source_fp (str): path to the file that must be converted.

        Returns:
            None
        """
        if self.VERBOSE:
            print("converting", source_fp)
        if dest_fp == None:
            dest_fp = self.make_dest_fp(source_fp)
        if not self.OVERWRITE and os.path.exists(dest_fp):
            print("Not overwriting already existing file:", dest_fp)
            print("Set self.overwrite to True if you want to overwrite")
            return
        
        shamid = re.findall("\d+", source_fp)[-1]

        book_data, toc_data, metadata = self.get_data(source_fp, shamid)

        metadata = self.format_metadata(metadata)

        text, notes = self.format_text(book_data, toc_data, shamid)
        text = self.reflow(text)
        #text = self.add_milestones(text)
        text = self.post_process(text)
        notes = self.reflow(notes)
        notes = re.sub("(\n+)(?![\n~P])", r"\1# ", notes)
        text = self.compose(metadata, text, notes)
        #print(9, text)

        self.save_file(text, dest_fp)
        #print("saved text to", dest_fp)

    def convert_files_in_folder(self, source_folder, dest_folder=None,
                                extensions=[], fn_regex=None):
        if "mdb" in " ".join(extensions):
            json_folder = self.mdb2json(source_folder, fn_regex=fn_regex)
            print("conversion from mdb to json completed.")
            extensions = ["json"]
        else:
            json_folder = os.path.join(self.dest_folder, "json")
        source_folder = json_folder
        self.dest_folder = os.path.join(self.dest_folder, "txt")
        print("Converting from json to txt...")
        super().convert_files_in_folder(source_folder, extensions=extensions,
                                        fn_regex=fn_regex)

    def mdb2json(self, source_folder, fn_regex=None):
        """Convert mdb files to json files.

        The .mdb files are usually in the Books folder,
        in subfolders numbered "0" to "9".
        The converted files will be saved in <self.dest_folder>/json/"""
        dest_folder = os.path.join(self.dest_folder, "json")
        if not os.path.exists(dest_folder):
            os.makedirs(dest_folder)
        mdb_files = []
        print("source folder:", source_folder)
        for root, dirs, files in os.walk(source_folder):
            for fn in files:
                if fn_regex:
                    if fn.endswith(".mdb") and re.findall(fn_regex, fn):
                        mdb_files.append(os.path.join(root, fn))
                else:
                    if fn.endswith(".mdb"):
                        mdb_files.append(os.path.join(root, fn))
        print("Converting {} .mdb files to json".format(len(mdb_files)))
        failed = []
        for fp in mdb_files:
            shamid = re.findall("\d+", fp)[-1]
            outfp = os.path.join(dest_folder, shamid+".json")

            print("connecting to db {}...".format(fp))
            conn, cur = bok.connect_to_db(fp)

            print("converting db to dict...")
            d = bok.mdb2dict(cur, VERBOSE=False)
            #print(d.keys())
            #input()

            print("saving to json...")
            if self.multiple_books_in_mdb:
                for k in d:
                    if k.startswith("b"):
                        shamid = k[1:]
                        outfp = os.path.join(dest_folder, shamid+".json")
                        try:
                            t = d["t"+shamid]
                        except:
                            print(k, "TITLE TABLE MISSING!")
                            t = {}
                        try:
                            b = d["b"+shamid]
                        except Exception as e:
                            print(k)
                            print(e)
                            b = {}
                        if b:
                            bok.save_to_json({"b"+shamid: b,
                                              "t"+shamid: t},
                                             outfp)
                        else:
                            print("CONVERSION FAILED")
                            failed.append(shamid)
##                        try:
##                            bok.save_to_json({"b"+shamid: d["b"+shamid],
##                                              "t"+shamid: d["t"+shamid]},
##                                             outfp)
##                        except:
##                            print(k)
##                            print("TITLE TABLE MISSING!")
##                            try:
##                                bok.save_to_json({"b"+shamid: d["b"+shamid],
##                                                  "t"+shamid: {}},
##                                                 outfp)
##                            except:
##                                print("CONVERSION FAILED")
##                                failed.append(shamid)
            else:
                try:
                    bok.save_to_json({"b"+shamid: d["book"],
                                      "t"+shamid: d["title"]},
                                     outfp)
                except:
                    print("TITLE TABLE MISSING!")
                    try:
                        bok.save_to_json({"b"+shamid: d["book"],
                                          "t"+shamid: {}},
                                         outfp)
                    except:
                        print("CONVERSION FAILED")
                        failed.append(fp)
                print("finished")
        if failed:
            print("Conversion of these {} files failed:".format(len(failed)))
            for f in failed:
                print("    ", f)
            input("Press Enter to continue")
        else:
            if VERBOSE:
                msg = "all {} mdb files converted ".format(len(mdb_files))
                print(msg + "successfully to", dest_folder)
        return dest_folder

    def get_data(self, source_fp, shamid):
        """Extract the data from the source_fp and format it in OpenITI format.

        ### Overwrite this method in subclass ###

        Args:
            source_fp (str): path to the original file

        Returns:
            text (str): the text in its initial state
        """
        with open(source_fp, mode="r", encoding="utf-8") as file:
            data = json.load(file)

        book_data = data["b"+shamid]

        # load the table of contents:
        toc_data = data["t"+shamid]

        # get the book's metadata.
        # For data converted from bok files,
        # extract it from the "Main" dictionary in the data dict;
        # If extracted from the Shamela program,
        # extract it from the dictionary that
        # contains metadata of all the books:

        if self.all_meta:
            book_meta = self.all_meta[shamid]
        else:
            book_meta = data["Main"]

        return book_data, toc_data, book_meta

    def format_text(self, book, toc, book_id):
        """Extract text from json, add page numbers and headings.

        Args:
            book (list): a list of  dictionaries
                (keys: id, nass, page, part, [seal])
            toc (list): a list of dictionaries
                (keys: id, tit, lvl, [sub])

        Returns:
            text (str): the formatted text
        """
        def make_numeric(n):
            if str(n).isnumeric():
                return int(n)
            else:
                return 0

        def format_struct_dict(toc):
            """turn the table of contents list into a dictionary.

            Args:
                toc (list): a list of dictionaries
                    (keys: id, tit, lvl, [sub])

            Returns:
                struct_dict (dict): key: id (i.e., the text fragment id;
                    value: list dictionaries that have the same id
                    (keys: id, tit, level, sub)
            """
            struct_dict = defaultdict(list)
            for row in toc:
                struct_dict[row["id"]].append(row)
            return struct_dict

        self.struct_dict = format_struct_dict(toc)
        #text = ["### |EDITOR|\n", ]
        text = []
        notes = []
        page_no = 0
        vol_page = ""
        try:
            book = sorted(book, key=lambda k: int(k["id"]))
        except Exception as e:
            print("book no.", book_id, ": unable to sort pages by id:", e)
            print("Attempting to sort by volume and page number...")
            try:
                book = sorted(book, key=lambda k: (int(k["part"]), int(k["page"])))
                print("successful")
            except:
                print("book no.", book_id, ": unable to sort pages by volume and page:", e)
                print("Keep book in the current order.")

        for row in book:
            # set variables for current row:

            if VERBOSE:
                print("vol {} page {} id {} : {}".format(row["part"],
                                                         row["page"],
                                                         row["id"],
                                                         row["nass"][:30]))
            passage_text = row["nass"]
            passage_id = row["id"]
            page_no = make_numeric(row["page"])
            vol_no = make_numeric(row["part"])
            vol_page = "\n\nPageV{:02d}P{:03d}\n\n".format(vol_no, page_no)



            # remove the short vowels etc.:

            passage_text = self.pre_process(passage_text)

            # remove the footnotes:

            passage_text, notes = self.remove_notes(passage_text, notes, vol_no, page_no)

            # add the structural formatting:

            if passage_id in self.struct_dict:
                passage_text = self.add_structural_annotations(passage_text,
                                                               passage_id)
            passage_text = re.sub("[\n\r¶]+ *(?![#\|P])", "\n\n# ", passage_text)

            if VERBOSE:
                print("vol {} page {} id {} : {}".format(vol_no,
                                                         page_no,
                                                         passage_id,
                                                         passage_text[:100]))
                input()

            text.append(passage_text)

            # add page numbers:

            if row["page"] != None:
                text.append(vol_page)
            else:
                if row["nass"]:
                    text.append(row["nass"] + "\n\n\n----NO PAGE NO------\n\n\n\n")

        # compile text and endnote strings from lists:
        text = "".join(text)
        if notes:
            notes = "\n\n### |EDITOR|\nENDNOTES\n\n" + "".join(notes)
        else:
            notes = ""

        return text, notes

    def remove_notes(self, passage_text, notes, vol_no, page_no):
        """Remove footnotes from the passage_text and add them to notes list."""
        footnotes = re.findall(self.footnote_regex,
                               passage_text, flags=re.DOTALL)
        if footnotes:
            fmt = "\n\n{}\nPageV{:02d}P{:03d}"
            notes.append(fmt.format(footnotes[0], vol_no, page_no))
            passage_text = re.sub(self.footnote_regex, "",
                                  passage_text, flags=re.DOTALL)
        return passage_text, notes

    def add_structural_annotations(self, passage_text, passage_id):
        """Add structural tags to titles in passage_text."""
        #print("passage_id", passage_id)
        for el in sorted(self.struct_dict[passage_id],
                         key= lambda item: item["lvl"],
                         reverse=True):
            lvl = el["lvl"]
            try:
                tit = deNoise(el["tit"])
            except:
                tit = "[NO TITLE]"
            #print("title", tit)
            tit_regex = re.sub("(\w+)", r"W\1", tit)[1:]
            tit_regex = re.sub("\W+", r"", tit_regex)
            tit_regex = re.sub("W", r"\W*", tit_regex)
            tit_regex = ".*{}.*".format(tit_regex)
            #print("regex", tit_regex)
            if re.findall(tit_regex, passage_text):
                r = "\n\n### {} AUTO ".format(int(lvl)*"|")+r"\1\n\n"
                passage_text = re.sub("[\n\r¶]*({})".format(tit_regex),
                                      r, passage_text, count=1)
            else:
                if VERBOSE:
                    print("!!!title not inside text", tit)
                t = "\n\n### {} CHECK [{}]\n\n".format(int(lvl)*"|", tit)
                passage_text = t + passage_text
        #print(passage_text[:120])
        return passage_text

    def format_metadata(self, metadata):
        """Convert the metadata in the metadata dict to a string."""
        m = []
        fmt = "#META# {}: {}"
        for k,v in metadata.items():
            if v:
                # treat the "betaka" field different from the other fields.
                # It contains a multi-line string with a number of headings.
                # Store each of these headings with their value separately:

                if k == "betaka":
                    m.append(fmt.format("Shamela_short_metadata_record", ""))
                    betaka = re.split("[\n\r]+", v)
                    for b in betaka:
                        b = re.split(": ", b, maxsplit=1)
                        if len(b)>1:
                            m.append(fmt.format(b[0], b[1]))
                        else:
                            m.append(fmt.format("digitization_comments", b[0]))
                else:
                    #print(col_headers[i], field)
                    #print(re.sub(r"\.?\n+", ". ", str(field)))
                    m.append(fmt.format(k, re.sub(r"\.?[\r\n]+", ". ", str(v))))
        for k,v in self.additional_meta.items():
            m.append(fmt.format(k,v))
        metadata = "\n\n".join(m)
        return self.magic_value + metadata + self.header_splitter

    def pre_process(self, text):
        if not text:
            return ""
        text = re.sub(" ?¶ ?", "\n\n", text)
        text = super().pre_process(text)
        return text

    def post_process(self, text):
        """Carry out post-processing operations on the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            texgt (str): the text after post-processing operations.
        """
        # remove unwanted spaces:
        text = re.sub("  +", " ", text)
        text = re.sub("\n ", "\n", text)
        text = re.sub("\n# ?\n+", "\n", text)

        #print(text[:50000])

        # remove the new lines before page numbers:
        text = re.sub(r" *\n+ *(PageV\d+P\d+)\n+", r" \1\n~~", text)
        # restore new lines when the page number is preceded by a relevant character:
        text = re.sub(r"""([.:!؟|*"]|AUTO|CHECK) +(PageV\d+P\d+)\n~~""",
                      r"\1\n\n\2\n\n", text)
        # restore new lines when the page number is followed by ###:
        text = re.sub(r"""\s+(PageV\d+P\d+)\n~~\s*(\#+)""", r"\n\n\1\n\n\2", text)
        # double the new lines when the page number is preceded by a relevant character
        #text = re.sub("""(?<=[.:!؟|*"])\n+(PageV\d+P\d+)""", r"\n\n\1", text)

        text = re.sub("(PageV\d+P\d+)(\n\n+)(?!#)", r"\1\2# ", text)

        text = re.sub("\n{2,}", "\n", text)
        text = re.sub("[\r\n]+# *(?:[\r\n]+|\Z)", "\n", text)

        return text


################################################################################


def extract_metadata(files_folder, outfolder):
    """Extract the book metadata from the Files folder of the Shamela program.

    The metadata is located in two separate mdb files:
    
    * Files/main.mdb
    * Files/special.mdb

    The metadata is extracted into json files in the outfolder:
    
    * [outfolder]/main_metadata.json: full metadata from Files/main.mdb
      in json format (dictionary). Under key "0bok" it contains a
      list of metadata dictionaries, one for each book in the collection
    * [outfolder]/special_metadata.json: full metadata from Files/special.mdb
      in json format
    * [outfolder]/genres.json: a dictionary of genres and their ids
      extracted from main.mdb
    * [outfolder]/Authors_metadata.json: a dictionary of authors and their ids
      extracted from special.mdb
    * [outfolder]/main_metadata_reform.json: a reformatted version of
      main_metadata.json (dictionary of dictionaries, one for each book:
      key = bookID, value = book metadata in dictionary format)

    Only main_metadata_reform.json is necessary for the conversion
    of Shamela mdb files to txt.
    """
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)

    # extract metadata from special.mdb (most importantly, the author metadata):
    fp = os.path.join(files_folder, "special.mdb")
    conn, cur = bok.connect_to_db(fp)
    d = bok.mdb2dict(cur, VERBOSE=False)
    print("Saving special.mdb to json...")
    bok.save_to_json(d, os.path.join(outfolder, "special_metadata.json"))
    print("Saving authors_dict...")
    authors_dict = {row["authid"]: row for row in d["Auth"]}
    bok.save_to_json(authors_dict,
                     os.path.join(outfolder, "Authors_metadata.json"))
    conn.close()

    # extract metadata from main.mdb:
    fp = os.path.join(files_folder, "main.mdb")
    conn, cur = bok.connect_to_db(fp)
    d = bok.mdb2dict(cur, VERBOSE=False)
    bok.save_to_json(d["0cat"], os.path.join(outfolder, "genres.json"))
    reformatted_list = reformat_metadata(d, authors_dict)
    d = {"0bok" : reformatted_list}
    bok.save_to_json(d, os.path.join(outfolder, "main_metadata.json"))
    conn.close()

    # save metadata also in dictionary format (key = bookID, value = data):
    new = dict()
    books = d["0bok"]
    for book_data in sorted(books, key=lambda x: int(x["bkid"])):
        #print(book_data["bkid"])
        if int(book_data["bkid"])%1000 == 0:
            print(book_data["bkid"])
        if book_data["bkid"] in new:
            print(book_data["bkid"], "double!")
        new[book_data["bkid"]] = book_data

    outfp = os.path.join(outfolder, "main_metadata_reform.json")
    with open(outfp, mode="w", encoding="utf-8") as file:
        json.dump(new, file, ensure_ascii=False, indent=4, sort_keys=True)




def reformat_metadata(d, authors_dict):
    """Reformats the Shamela Betaka

    The Shamela metadata table contains a field called 'betaka',
    i.e., a kind of human-readable index card of the main metadata
    of the book (author, title, etc.).
    The betaka field is a string, with key: value pairs split by
    new lines. This function splits the metadata in the
    betaka string and adds it to the meta dictionary.

    Args:
        d (dict): a dictionary containing two items:
            d["0bok"]: a list of dictionary representations of the
            metadata of each book in the Shamela metadata table 'Main'.
            d["0cat"]: a list of dictionary representations of the
            genre metadata in the Shamela metadata table
            (keys: id, name, catord, lvl)
    Returns:
        meta (list): the same list of dictionaries, but with the metadata
            in the betaka field split out into keys and values
    """
    genres = {row["id"] : row for row in d["0cat"]}
    meta = d["0bok"]

    for i, row in enumerate(meta):

        # reformat the betaka:

        if row["betaka"]:
            meta = reformat_betaka(meta, i, row)

        # replace the genre id with the genre name:

        if row["cat"]:
            id_ = row["cat"]
            meta[i]["cat"] = genres[id_]["name"]

        if row["authno"]:
            id_ = row["authno"]
            for k,v in {"auth":"auth",
                        "higrid":"higrid",
                        "ad":"ad",
                        "Lng":"lng",
                        "authinf":"inf"}.items():
                if authors_dict[id_][v]:
                    if k in meta[i]:
                        if meta[i][k]:
                            if authors_dict[id_][v] not in meta[i][k]:
                                meta[i][k] += " :: " + authors_dict[id_][v]
                        else:
                            meta[i][k] = authors_dict[id_][v]
                    else:
                        meta[i][k] = authors_dict[id_][v]
    return meta


def reformat_betaka(meta, i, row):
    """Reformats the Shamela Betaka

    The Shamela metadata table contains a field called 'betaka',
    i.e., a kind of human-readable index card of the main metadata
    of the book (author, title, etc.).
    The betaka field is a string, with key: value pairs split by
    new lines. This function splits the metadata in the
    betaka string and adds it to the meta dictionary.
    """

    betaka = re.split("[\n\r¶]+", row["betaka"])
    for b in betaka:
        b = re.split(": ", b, maxsplit=1)
        if len(b)>1:
            meta[i][b[0].strip()] = b[1].strip()
        else:
            meta[i]["digitization_comments"] = b[0].strip()
    return meta


def print_default_config_file():
    config="""\
# Shamela converter configuration file:

# path to the Files folder of the shamela database
# (this folder contains the main metadata database files):
files_folder = None

# path to the Books folder of the shamela database:
# (this folder contains subfolders with the separate book .mdb files)
books_folder = None

# path to the folder that contains the converted metadata files
# (if the metadata has been extracted before):
meta_folder = None

# path to the folder in which the converted text files should be saved:
conv_folder = None

# date when the database was downloaded:
download_date = ""

# name and/or url of the downloaded database:
download_source = ""

# regular expression pattern for file names to be converted:
fn_regex = None

# extensions of the files that should be converted:
extensions = ["mdb"]

"""
    print(config)



if __name__ == "__main__":
    #print_default_config_file()
    # path to the Files folder of the shamela database
    # (this folder contains the main metadata database files):
    files_folder = r"G:\London\OpenITI\new\Zaydiyya\Zaydiyya\Files"

    # path to the Books folder of the shamela database:
    # (this folder contains subfolders with the separate book .mdb files)
    books_folder = r"G:\London\OpenITI\new\Zaydiyya\Zaydiyya\Books"

    # path to the folder that contains the converted metadata files
    # (if the metadata has been extracted before):
    meta_folder = r"G:\London\OpenITI\new\Zaydiyya\Zaydiyya\extracted_metadata"

    # path to the folder in which the converted text files should be saved:
    conv_folder = r"G:\London\OpenITI\new\Zaydiyya\Zaydiyya\BooksConverted2"

    # date when the database was downloaded:
    download_date = "2020-04-16"

    # name and/or url of the downloaded database:
    download_source = "al-Maktaba al-Shamela al-Zaydiyya"

    # regular expression pattern for file names to be converted:
    fn_regex = "337"

    # extensions of the files that should be converted:
    extensions = ["mdb"]

    main(files_folder=files_folder,
         books_folder=books_folder,
         meta_folder=meta_folder,
         conv_folder=conv_folder,
         download_date=download_date,
         download_source=download_source,
         fn_regex=fn_regex,
         extensions=extensions)



"""Convert Epubs from the Masaha Hurra library to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single epub file.
* convert_files_in_folder: convert all epub files in a given folder

Usage examples:
    >>> folder = "test/masaha/epub/"
    >>> meta_fp = "test/masaha/all_books/meta/all_metadata.json"
    >>> from epub_converter_masaha import convert_file, convert_files_in_folder
    >>> src_fp = folder+"000008.epub"
    >>> convert_file(src_fp, meta_fp, dest_fp=folder+"converted/Masaha000008")
    >>> convert_files_in_folder(folder, meta_fp, dest_folder=folder+"converted")
    Converting all files in folder test/masaha/epub with extensions ['epub']

Both functions use the MasahaEpubConverter class to do the heavy lifting.
The MasahaEpubConverter is a subclass of the GenericEpubConverter,
which in turn is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    \_ GenericEpubConverter
            \_ MasahaEpubConverter

Methods of both classes:

(methods of GenericConverter are inherited by GenericEpubConverter;
methods of GenericConverter with the same name
in GenericEpubConverter are overwritten by the latter)

=========================== ========================= =======================
generic_converter           epub_converter_generic    epub_converter_masaha 
=========================== ========================= =======================
__init__                    __init__                  __init__ 
convert_files_in_folder     (inherited)               (inherited)
convert file                (inherited)               (inherited)
make_dest_fp                (inherited - generic!)    (inherited - generic!)
get_metadata                (inherited - generic!)    get_metadata
get_data                    get_data                  (inherited)
pre_process                 (inherited)               (inherited)
add_page_numbers            (inherited - generic!)    (inherited - generic!)
add_structural_annotations  (inherited - generic!)    (inherited - generic!) 
remove_notes                remove_notes              (inherited)
reflow                      (inherited)               (inherited)
add_milestones              (inherited)               (inherited)
post_process                (inherited - generic!)    post_process
compose                     (inherited)               (inherited)
save_file                   (inherited)               (inherited)
                            convert_html2md           convert_html2md
                            inspect_epub              (inherited)
                            sort_html_files_by_toc    sort_html_files_by_toc
                            add_unique_tags           (inherited)
=========================== ========================= =======================


Examples:
    >>> from epub_converter_masaha import MasahaEpubConverter
    >>> from helper.yml2json import yml2json
    >>> folder = "test/"
    >>> fn = "26362727.epub"
    >>> hc = MasahaEpubConverter(dest_folder="test/converted")
    >>> hc.VERBOSE = False
    >>> meta_fp = "masaha/all_books/meta/all_metadata.json"
    >>> hc.metadata_file = meta_fp
    >>> hc.convert_file(folder+fn)

    #>>> hc.convert_files_in_folder(folder)

"""

import os
import json
import shutil
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.epub_converter_generic import GenericEpubConverter
from openiti.new_books.convert.helper import html2md_masaha
from openiti.new_books.convert.helper.yml2json import yml2json


def convert_file(fp, meta_fp, dest_fp=None, verbose=False, overwrite=False):
    """Convert one file to OpenITI format.

    Args:
        fp (str): path to the file that must be converted.
        meta_fp (str): path to the yml file containing the Masaha metadata
        dest_fp (str): path to the converted file.

    Returns:
        None
    """
    conv = MasahaEpubConverter(overwrite=overwrite)
    conv.VERBOSE = verbose
    with open(meta_fp, mode="r", encoding="utf-8") as file:
        d = json.load(file)
        conv.metadata_dic = {int(item["book_id"]): item for item in d}
    conv.metadata_file = meta_fp
    conv.convert_file(fp, dest_fp=dest_fp)

##def convert_multifile_text(folder, meta_fp, dest_folder, verbose=False):
##    for i, fn in enumerate(os.listdir(folder)):
##        if i == 0:
##            dest_fp = os.path.join(dest_folder, os.path.splitext(fn)[0])
##        
        

def convert_files_in_folder(src_folder, meta_fp, dest_folder=None, verbose=False,
                            extensions=["epub"], exclude_extensions=["yml"],
                            fn_regex=None, overwrite=False):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        meta_fp (str): path to the yml file containing the Masaha metadata
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    msg = "Converting all files in folder {} with extensions {}"
    print(msg.format(src_folder, extensions))
    conv = MasahaEpubConverter(overwrite=overwrite)
    conv.VERBOSE = verbose
    with open(meta_fp, mode="r", encoding="utf-8") as file:
        d = json.load(file)
        conv.metadata_dic = {int(item["book_id"]): item for item in d}
    conv.metadata_file = meta_fp
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################




class MasahaEpubConverter(GenericEpubConverter):
    def __init__(self, dest_folder=None, overwrite=True):
        super().__init__(dest_folder=dest_folder, overwrite=overwrite)
        self.toc_fn = "content.opf"
        self.metadata_file = None


    def convert_files_in_folder(self, source_folder, dest_folder=None,
                                extensions=[], exclude_extensions=[],
                                fn_regex=None):
        """Convert all files in a folder to OpenITI format.\
        Use the `extensions` and `exclude_extensions` lists to filter\
        the files to be converted.

        Args:
            source_folder (str): path to the folder that contains
                the files that must be converted.
            extensions (list): list of extensions; if this list is not empty,
                only files with an extension in the list should be converted.
            exclude_extensions (list): list of extensions;
                if this list is not empty,
                only files whose extension is not in the list will be converted.
            fn_regex (str): regular expression defining the filename pattern
                e.g., "-(ara|per)\d". If `fn_regex` is defined,
                only files whose filename matches the pattern will be converted.

        Returns:
            None
        """
        failed = []
        if dest_folder:
            self.dest_folder = dest_folder
        fp_list = self.filter_files_in_folder(source_folder, extensions,
                                              exclude_extensions, fn_regex)
        for fp in fp_list:
            print(fp)
            try:
                self.convert_file(fp)
            except Exception as e:
                print("ERROR:", e)
                failed.append((fp, e))
                

        # deal with multivolume texts that are in separate folders:
        multivol_folders = [f for f in os.listdir(source_folder) if f.startswith("multivol")]
        multivol_folders = [os.path.join(source_folder, f) for f in multivol_folders]
        
        for folder in multivol_folders:
            print(folder)
            try:
                first_fn = sorted(os.listdir(folder))[0]
            except Exception as e:
                print("folder does not contain files:", e)
                failed.append((folder, e))
                continue
            outfn = re.sub("\.epub", "Vols.automARkdown", first_fn)
            outfp = os.path.join(dest_folder, outfn)
            if os.path.exists(outfp):
                print(outfp, "already exists")
                continue
            temp_folder = os.path.join(dest_folder, "temp")
            if os.path.exists(temp_folder):
                shutil.rmtree(temp_folder)
            os.makedirs(temp_folder)
            self.convert_files_in_folder(folder, dest_folder=temp_folder,
                                         extensions=extensions,
                                         exclude_extensions=exclude_extensions,
                                         fn_regex=fn_regex)
            # combine all volumes into one folder:
            combined = []
            endnotes = []
            for i, fn in enumerate(sorted(os.listdir(temp_folder))):
                fp = os.path.join(temp_folder, fn)
                if i == 0:
                    outfn = re.sub("\.", "Vols.", fn) 
                    outfp = os.path.join(dest_folder, outfn)
                with open(fp, mode="r", encoding="utf-8") as file:
                    text = file.read()
                    if i != 0:
                        text = re.split("#META#Header#End#?", text)[-1]
                        page = "PageV{:02d}P{:03d}"
                        text = re.sub("PageV\d+P(\d+)", r"PageV{:02d}P\1".format(i+1), text)
                if re.findall("### \|EDITOR\|[ \r\n]+ENDNOTES:?", text):
                    text, notes = re.split("### \|EDITOR\|[ \r\n]+ENDNOTES:?", text)
                    endnotes.append(notes)
                combined.append(text)
            with open(outfp, mode="w", encoding="utf-8") as file:
                text = "\n\n".join(combined)
                endnotes = "\n\n".join(endnotes)
                file.write(text + "\n\n### |EDITOR\|\n\nENDNOTES:\n\n" + endnotes)
        print("Converting all files done")
        if failed:
            print("These files failed to convert:")
            for fp, e in failed:
                print(fp, e)
        
                

    def sort_html_files_by_toc(self, zp, toc_fp, html_files):
        """Gets the table of contents from the Epub file.

        Args:
            zp: zipfile object
            toc_fp (str): filepath to the table of contents of the epub.
            html_files(list): an unordered list of the html files in the epub.

        Returns:
            toc (list): a list of filepaths to the html files
                in the epub file, in the order specified by the
                table of contents
        """
        html_files_dict = {os.path.split(fp)[-1] : fp for fp in html_files}
        toc_data = zp.read(toc_fp)
        toc_data = codecs.decode(toc_data, "utf-8")
        soup = BeautifulSoup(toc_data)
        toc_ol = soup.find("spine")
        toc = []
        for item in toc_ol.find_all("itemref"):
            fn = os.path.split(item.get("idref"))[-1]
            if fn in html_files_dict:
                toc.append(html_files_dict[fn])
        return toc

    def convert_html2md(self, html):
        """Use custom html to mARKdown function for Masaha epubs."""
        text = html2md_masaha.markdownify(html)
        return text

    def get_metadata(self, metadata_fp):
        """Custom method to get the metadata of the Masaha epub file."""
        source_fp = self.source_fp
        bookID = os.path.split(source_fp)[1]
        bookID = int(os.path.splitext(bookID)[0])
        meta_dic = self.metadata_dic[bookID]
        meta = ["#META# {}: {}".format(k,v) for k,v in sorted(meta_dic.items())]
        return self.magic_value + "\n".join(meta) + self.header_splitter

    def post_process(self, text):
        """Custom post-processing for masaha texts"""
        # put page number at the bottom of the page:
        text = re.sub("(PageV\d+P\d+)(.+?)(?=Page|\Z)", r"\2\n\n\1\n\n", text, flags=re.DOTALL)
        processed = super().post_process(text)
        return processed
        


if __name__== "__main__":
    #import doctest
    #doctest.testmod()
    #input("Testing finished. Continue?")

    # identify the location of the yml file containing the metadata:
    meta_fp = r"test\masaha\meta\all_metadata.json"
    src_folder = "test/masaha/epub"
    convert_files_in_folder(src_folder, meta_fp, dest_folder="test/converted", verbose=False)
##    hc.metadata = yml2json(meta_fp, container={})
    
##    fp = r"test\26362727.epub"
##    hc.convert_file(fp)
##    print("converted Masaha epub", fp)
##
##    hc.convert_files_in_folder("test/masaha")
##    print("converted all epub files in folder", "test/masaha")




"""Clone github repositories to your local machine.

Examples:
    Clone all OpenITI 25-year repos::
    
        # >>> repo_list = get_repo_urls(group="orgs", name="OpenITI", 
        #                                path_pattern="\\d{4}AH$")
        # >>> clone_repos(repo_list, r"D:\\OpenITI")

    Clone specific repos::
    
        # >>> base = ""https://github.com/OpenITI/"
        # >>> repo_list = [base+"mARkdown_scheme", base+"RELEASE"]
        # >>> clone_repos(repo_list, r"D:\\OpenITI")

Command line usage:
    To clone all the OpenITI organization's 25-years repos:
    
        ``python clone_OpenITI.py ["orgs" or "users"] [user or org name] [path_pattern] [[dest_folder]]``

        ``dest_folder$ python pth/to/clone_OpenITI.py orgs OpenITI \d+AH$``
    
        ``other_folder$ python pth/to/clone_OpenITI.py orgs OpenITI \d+AH$ path/to/dest_folder``

    If fewer than 3 arguments are given, the program will prompt you for the arguments::

        $ python clone_OpenITI.py
        Enter the organization/user name: openiti
        Organization or User? (orgs/users): orgs
        Enter the regex pattern to match desired repo URLs: \d{4}AH
        Enter the destination folder for the clone: openiti_clone
"""

import sys
import requests
import re
import subprocess


def get_repo_urls(group="orgs", name="OpenITI", path_pattern=r"\d{4}AH$"):
    """Get a list of all repos owned by organisation/user `name`
    that match the regex `path_pattern`

    Args:
        group (str): either "users" or "orgs". Defaults to "orgs"
        name (str): GitHub name of the user/organization. Defaults to "OpenITI"
        path_pattern(str): regex pattern that matches the desired repository names.
            If none is defined, all repos will be cloned. Defaults to r"\d{4}AH$".

    Returns:
        (list): a list of repo urls that matches the `path_pattern` regex.
    """
    repo_urls = []
    page = 1
    while True:
        url = 'https://api.github.com/{0}/{1}/repos?per_page=100&page={2}'
        r = requests.get(url.format(group, name, page))
        print(url.format(group, name, page))
        if r.status_code == 200:
            rdata = r.json()
            for repo in rdata:
                if path_pattern:
                    if re.compile(path_pattern).search(repo['full_name']):
                        repo_urls.append(repo['clone_url'])
                else:
                    repo_urls.append(repo['clone_url'])
            if (len(rdata) >= 100):
                page = page + 1
            else:
                print('Found {0} repos.'.format(len(repo_urls)))
                break
        else:
            print(r)
            return False
    return repo_urls


def clone_repos(all_repos, clone_dir="."):
    """Clone the list of repo urls `all_repos` to the `clone_dir`

    Args:
        all_repos (list): a list of repo urls.
        clone_dir (str): path to the local folder where the repos are to be cloned.
            Defaults to the current active directory.

    Returns:
        None
    """
    print('Cloning...')
    counter = 1
    total_repo_nr = str(len(all_repos))
    for repo in all_repos:
        path_to_clone = clone_dir #+ repo.split("/")[-1]
        print ('\nRepository ' + str(counter) + ' of ' + total_repo_nr + '\n')
        print(path_to_clone)
        p = subprocess.Popen(["git", "clone", repo], cwd=path_to_clone)
        p.wait()
        # os.system('git clone ' + repo + ' ' + path_to_clone + repo.split("/")[-1])
        counter += 1


if __name__ == '__main__':

    # path_pattern = re.compile("\d{4}AH$")
    if len(sys.argv) < 4:
        name = input("Enter the organization/user name: ")
        group = input("Organization or User? (orgs/users): ")
        path_pattern = input("Enter the regex pattern in repo URL: ")
        clone_dir = input("Enter the destination folder for the clone: ")
    else:
        name = sys.argv[1]
        group = sys.argv[2]
        path_pattern = sys.argv[3]
        if sys.argv < 5:
            clone_dir = "."
        else:
            clone_dir = sys.argv[4]

    # print("arg: " ,sys.argv)
    if len(sys.argv) > 0:
        total = get_repo_urls(group, name, path_pattern)
        if total:
            clone_repos(total, clone_dir)
        else:
            print("No repository found!")
    else:
        print('In Terminal, cd to the destination directory in which you want to clone the repos.\n '
              'then run the script:\n'
              '\tpython3 path/to/script/clone_OpenITI.py OpenITI\n'
              'OpenITI is the organization name!')



from git import Repo
import os


class GitUtil(object):

    def __init__(self, repo: str, user: dict, path: str = ''):
        """ initial method
        To use this class, first make sure you have set ssh private key and give the path
        in the following wapper.
        Save the wrapper as a bash script and give required permissions using "sudo chmod a+x /path/to/ssh_key":

        #!/bin/bash
        ssh -i path/to/ssh/key -oIdentitiesOnly=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null "$@"

        Parameters
        ----------
        repo : str
            The repository ssh url in `git@` format.
        user : dict
            User information: `ssh` wrapper script path, github username and email in the following format, e.g:
            {
                'ssh': '/usr/local/bin/ssh_wrapper',
                'name': 'name',
                'email': 'name@mail.com'
            }
        path : str, optional
            The checkout path.
            If you didn't specify this value, it will based on your current working directory.
            **Make sure that's an empty folder for `git clone`**

            Examples
        --------
        from .utils.git import GitUtils
        git = GitUtils(
            repo='git@github.com:username/repo.git',
            user={
                'ssh': '/usr/local/bin/ssh_wrapper.sh   ',
                'name': 'username',
                'email': 'email@gmail.com',
            }
        )
        # change some file(s) here
        # add, commit, and push files to master
        git.commit('master','commit msg')
        """

        repo = repo.strip()
        path = os.path.join(os.getcwd(), path.strip())
        _ = {'ssh': '', 'name': 'name', 'email': 'name@gmail.com'}
        _.update(user)
        user = _

        if not repo.startswith('git@'):
            raise Exception(
                f'Invalid checkout url: {repo}\n\n'
                'Please use the valid ssh url with the following format:\n'
                ' `git@github.com:account/repository.git` format\n\n'
            )

        if not os.path.isfile(user['ssh']):
            raise Exception(
                f'Missing custom ssh wrapper {user["ssh"]}!\n\n'
                'Please provide a custom shh wrapper script with the correct ssh private key.\n'
                'The bash script should contain this line:\n\n'
                'ssh -i <path_to_ssh_private_key> -oIdentitiesOnly=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null "$@"\n\n'
            )

        os.environ['GIT_SSH'] = user['ssh']
        if os.path.isdir(path):
            self.repo = Repo(path)
            self.repo.git.pull('origin', 'master')
        else:
            os.makedirs(path)
            self.repo = Repo.clone_from(repo, path, env={'GIT_SSH': user['ssh']})
            self.repo.config_writer().set_value('user', 'name', user['name']).release()
            self.repo.config_writer().set_value('user', 'email', user['email']).release()

    def commit(self, branch: str = 'master', message: str = 'Auto commit'):
        """Commit method.
        This method will add, commit, and push all changes including untracked file.

        Parameters
        ----------
        branch : str, optional
            The branch you would like to commit. Default is master
        message : str, optionl
            Commit message
        """
        has_changed = False

        # Check if there's any untracked files
        for file in self.repo.untracked_files:
            print(f'Added untracked file: {file}')
            self.repo.git.add(file)
            if has_changed is False:
                has_changed = True

        if self.repo.is_dirty() is True: # reacts like git status
            for file in self.repo.git.diff(None, name_only=True).split('\n'):
                print(f'Added file: {file}')
                self.repo.git.add(file)
                if has_changed is False:
                    has_changed = True

        if has_changed is True:
            self.repo.git.commit('-m', message)
            self.repo.git.push('origin', branch)



"""Functions to read and write yaml files.

OpenITI metadata is stored in yaml files.
(yaml stands for "yet another markup language")

NB: In correctly formatted OpenITI yml files,
    - keys (lemmata) should always:
        * contain at least one hash (#)
        * end with a colon (:)
        * be free of any other non-letter/numeric characters

    - Values:
        * may contain any character, including colons
          and new line characters
          (only something that looks like a yml key
          (i.e., a combination of letters and hashes ending with a colon)
          should be avoided at the beginning of a line)
        * mutiline values should be indented with 4 spaces

The ymlToDic and dicToYML functions will retain double new lines
and new lines before bullet lists (in which bullets are `*` or `-`)
"""

import re
import textwrap

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    sys.path.append(root_folder)

from openiti.helper.templates import author_yml_template, book_yml_template, \
                                     version_yml_template



def ymlToDic(yml_str, reflow=False, yml_fp=""):
    """Convert a yml string into a dictionary.

    NB: in order to be read correctly, OpenITI yml keys (lemmata) should always:
        * contain at least one hash (#)
        * end with a colon (:)
        * be free of any other non-letter/numeric characters

        Values may contain any character, including colons.

        In multiline values, every new line should be indented with 4 spaces;
        multiline values may use double new lines and
        bullet lists (using `*` or `-` for items) for clarity.

    Args:
        yml_str (str): a yml string.
        reflow (bool): if set to False, the original layout
            (line endings, indentation) of the yml file
            will be preserved (useful for files containing bullet lists etc.);
            in the output string, new line characters will be replaced with ¶.
            if set to True, new line characters will be removed
            (except double line breaks and line breaks in bullet lists)
            and the indentation and line length will be standardized.

    Returns:
        (dict): dictionary representation of the yml key-value pairs

    Examples:
        >>> from yml import ymlToDic
        >>> yml_str = "\
00#BOOK#URI######: 0845Maqrizi.Muqaffa\\n\
90#BOOK#COMMENT##: multiline value; presence\\n\
    of colons: not a problem\\n\\n\\n\
".replace("        ", "") # remove Python indentation for doctest
        >>> yml_dic = {'00#BOOK#URI######:': '0845Maqrizi.Muqaffa',\
                       '90#BOOK#COMMENT##:': 'multiline value; presence of colons: not a problem'}
        >>> ymlToDic(yml_str, reflow=True) == yml_dic
        True
        
    """
    if yml_str.strip() == "":
        #raise Exception(yml_fp, "Empty YML file!")
        return {}
    
    # normalize new line characters:
    data = re.sub("\r\n", "\n", yml_str)
    
    # remove empty lines and spaces at end and beginning of string:
    data = data.strip() 

    # keep empty lines in multiline values:
    data = re.sub("\n([ \t]*)\n+([ \t]+)", r"¶\2¶\2", data)

    # keep linebreaks before bullet list items in multiline values:
    data = re.sub(r"[\n¶]([ \t]+[\*\-])", r"¶\1", data)
    
    if reflow: # remove other line breaks:
        data = re.sub(r"-\n+[ \t]+", "-", data)
        data = re.sub(r"\n+[ \t]+", " ", data)
    else: # keep line breaks:
        data = re.sub(r"\n([ \t]+)", r"¶\1", data)

    # split into key-value pairs and convert to dictionary:

    data = data.split("\n")
    dic = dict()
    for d in data:
        spl = re.split(r"^((?:#+\w+|\w+#+)[\w#]*:+)", d, 1)
        try:
            dic[spl[1]] = spl[2].strip()
        except:
            raise Exception(yml_fp, "no valid yml key in line", d)

    return dic


def readYML(fp, reflow=False):
    """Read a yml file and convert it into a dictionary.

    Args:
        fp (str): path to the yml file.
        reflow (bool): if set to False, the original layout
            (line endings, indentation) of the yml file
            will be preserved (useful for files containing bullet lists etc.);
            in the output string, new line characters will be replaced with ¶.
            if set to True, new line characters will be removed
            (except double line breaks and line breaks in bullet lists)
            and the indentation and line length will be standardized.
            
    Returns:
        (dict): dictionary representation of the yml key-value pairs

    Examples:
##        >>> fp = "D:/London/OpenITI/25Y_repos/0450AH/data/0429AbuMansurThacalibi/0429AbuMansurThacalibi.AhsanMaSamictu/0429AbuMansurThacalibi.AhsanMaSamictu.Shamela0025011-ara1.yml"
##        >>> readYML(fp)
##        {}
    """
    with open(fp, "r", encoding="utf-8") as file:
        s = file.read()
        try:
           return ymlToDic(s, yml_fp=fp)
        except Exception as e:
           print(fp)
           print(e)


def dicToYML(dic, max_length=80, reflow=True, break_long_words=False):
    """Convert a dictionary into a yml string.

    NB: use the pilcrow (¶) to force a line break within dictionary values.

    Args:
        dic (dict): a dictionary of key-value pairs.
        max_length(int): the maximum number of characters a line should contain.
        reflow (bool): if set to False, the original layout
            (line endings, indentation) of the yml string
            will be preserved (useful for files containing bullet lists etc.);
            if set to True, the indentation and line length
            will be standardized.
        break_long_words (bool): if False, long words will be kept on one line
            
    Returns:
        (str): yml string representation of the dic's key-value pairs

    Examples:
        >>> yml_dic = {'00#BOOK#URI######:': '0845Maqrizi.Muqaffa',\
                       '90#BOOK#COMMENT##:': 'multiline value; presence of colons: not a problem¶¶    * bullet point 1¶    * bullet point 2'}
        >>> yml_str = '\
        00#BOOK#URI######: 0845Maqrizi.Muqaffa\\n\
        90#BOOK#COMMENT##: multiline\\n\
            value; presence of colons: not\\n\
            a problem\\n\
        \\n\
            * bullet point 1\\n\
            * bullet point 2\
        '.replace("        ", "") # remove Python indentation for doctest
        >>> dicToYML(yml_dic, max_length=30, reflow=True) == yml_str
        True
        
        >>> yml_str = '\
        00#BOOK#URI######: 0845Maqrizi.Muqaffa\\n\
        90#BOOK#COMMENT##: multiline value; presence of colons: not a problem\\n\
            \\n\
            * bullet point 1\\n\
            * bullet point 2\
        '.replace("        ", "") # remove Python indentation for doctest
        >>> dicToYML(yml_dic, max_length=30, reflow=False) == yml_str
        True
    """
    data = []
    if dic:
        for k,v in dic.items():
            if k.strip().endswith(":"):
                i = "{} {}".format(k.strip(), str(v).strip())
            else:
                i = "{}: {}".format(k.strip(), str(v).strip())

            # split long values into indented multiline values:

            if "#URI#" not in i:

                lines = re.split("¶", i)

                if len(lines) > 1:
                    lines = [lines[0]] + [line if line.startswith((" ", "\t")) else "    "+line
                                          for line in lines[1:]]
                
                if reflow:
                    lines = ["\n    ".join(textwrap.wrap(line.strip(), max_length,
                                                         break_long_words=break_long_words))
                             for line in lines]

                i = "\n    ".join(lines)
            data.append(i)

    return "\n".join(sorted(data))

def fix_broken_yml(fp, execute=True):
    """Fix a yml file that is broken because
    (1) a line does not start with a valid key or space
    or (2) the colon after the key is absent

    Args:
        fp (str): path to the broken yml file
        execute (bool): if False, user's judgment about the fix
             will be asked before the fix is implemented

    Returns:
        None or yml_d
    """
    with open(fp, mode="r", encoding="utf-8") as file:
        data = file.read()
    key_lines = []
    current = []
    for line in data.splitlines():
        if re.findall("^\d\d#[#\w]{14}:?", line):
            if current:
                key_lines.append(current)
                current = []
        current.append(line.strip())
    if current:
        key_lines.append(current)
    key_lines = ["¶".join(line) for line in key_lines]
    key_regex = "^(\d\d#[#\w]{14}):?"
    yml_d = {re.findall(key_regex, line)[0]+":" : re.sub(key_regex, "", line).strip()
             for line in key_lines}
    print("original yml file:\n")
    print(data)
    print("\nAttempt to solve the issue:\n")
    print(dicToYML(yml_d))
    if execute or input("\nAccept change? Y/N: ").lower() == "y":
        return yml_d
    else:
        print("Aborting change. Please review YML file manually")
        return


if __name__ == "__main__":
    yml_fp = "../../../../../OpenITI/Github_clone/0875AH/data/0852IbnHajarCasqalani/0852IbnHajarCasqalani.InbaGhumr/0852IbnHajarCasqalani.InbaGhumr.Shamela0026317-ara1.yml" 
#    with open(yml_fp, mode="r", encoding="utf-8") as file:
#        t = file.read()
#    print(t)
#    input("continue?")
#    t = t.replace("\xa0", " ")
#    print(t)
#    input("continue?")
#    with open(yml_fp, mode="w", encoding="utf-8") as file:
#        file.write(t)
#    yml_dic = readYML(yml_fp)
#    print(yml_dic)
#    yml_str = dicToYML(yml_dic)
#    print(yml_str)
#    input("continue?")
    yml_str = """\
key1: short description
key2: longer description with : :: colons
key3: longer multiline description longer multiline description longer multiline description
    longer multiline description

    longer multiline description after double line break
    longer multiline description

    * bullet item 1
    * bullet item 2
"""
    print("do not reflow:")
    print(ymlToDic(yml_str, reflow=False))
    print(dicToYML(ymlToDic(yml_str, reflow=False), reflow=False, max_length=25))
    print("*******************")
    print("reflow:")
    print(ymlToDic(yml_str, reflow=True))
    print(dicToYML(ymlToDic(yml_str, reflow=True), reflow=True, max_length=25))
    d = {'key2:': 'longer description with : :: colons',
         'key1:': 'short description',
         'key3:': 'longer multiline description longer multiline description longer multiline description longer multiline description¶¶longer multiline description after double line break longer multiline description¶    ¶    * bullet item 1¶    * bullet item 2'}
    print("*******************")
    print(dicToYML(d, reflow=True, max_length=25))
    print("*******************")
    print(dicToYML(d, reflow=False, max_length=25))
    import doctest
    input()

#    doctest.testmod()
